# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from functools import cache
from typing import Callable, Iterable, Optional

from cachetools.func import ttl_cache

from more_itertools import seekable

from .client import JiraClient
from .converter import (
    JiraConverter
)
from .filter import (
    JiraFilter
)
from .sort import (
    JiraSorter
)
from ..core import (
    DataObject,
    DataSource,
    DataSourceError,
    DataSourceFilter
)
from ..core.operator import (
    DetailGetter,
    ListGetter,
    PageGetter,
    Relational
)
from ..core.relationship import RelationshipConfig

if typing.TYPE_CHECKING:
    from ..core.session import OperableSession

ClientFactory = Callable[[], JiraClient]
FilterFactory = Callable[[], JiraFilter]
SorterFactory = Callable[[], JiraSorter]
JiraConverterFactory = Callable[[], JiraConverter]


class JiraDataSource(
    DataSource,

    # the supported operators
    DetailGetter,
    ListGetter,
    PageGetter,
    Relational
):
    """
    A `DataSource` that connects to a remote JIRA

    Developers should likely use `create_jira_datasource`
    instead of this directly.
    """

    def __init__(
        self,
        client_factory: ClientFactory,
        jira_converter_factory: JiraConverterFactory,
        filter_factory: FilterFactory,
        sorter_factory: SorterFactory
    ) -> None:

        self.__client_factory = client_factory
        self.__jc_factory = jira_converter_factory
        self.__filter_factory = filter_factory
        self.__sorter_factory = sorter_factory
        super().__init__({})
        self.DEFAULT_PAGE_SIZE = 100

    @ttl_cache(ttl=60)
    def get_fields(self) -> dict:
        return self.__client_factory().get_fields()

    def __get_filter_string(
        self,
        object_filters: Optional[DataSourceFilter]
    ) -> Optional[str]:

        if object_filters is None:
            return ''
        return self.__filter_factory().dumps(object_filters)

    def __get_sort_string(
        self,
        sort_by: Optional[str]
    ) -> str:
        return self.__sorter_factory().sort(sort_by)

    @property
    @cache
    def attribute_types(self) -> dict[str, dict[str, str]]:
        fields = self.get_fields()
        return {
            'issue': {
                field['system_name']: field['type']
                for _, field in fields.items()
                if field['relation'] is None
            } | {
                'status_changes': 'List[Dict[str, Any]]'
            },
            'user': {
                'name': 'str',
                'emailAddress': 'str',
                'displayName': 'str'
            }
        }

    @property
    @cache
    def supported_types(self) -> list[str]:
        return list(self.attribute_types.keys())

    def get_by_id(
        self,
        object_type: str,
        object_ids: Iterable[str],
        **kwargs,
    ) -> Iterable[Optional[DataObject]]:
        if object_type not in self.supported_types:
            raise DataSourceError(f'{object_type} is not supported')

        client = self.__client_factory()
        jira_response = client.get_detail(object_type, object_ids)
        jira_converter = self.__jc_factory()

        converted_objects, _ = jira_converter.convert_list(jira_response) \
            if jira_response is not None else ([], 0)
        seekable_objects = seekable(converted_objects)
        for id_ in object_ids:
            seekable_objects.seek(0)
            for obj in seekable_objects:
                if obj.id == id_:
                    yield obj
                    break
            else:
                yield None

    def get_list_page(
        self,
        object_type: str,
        page_number: int,
        page_size: Optional[int] = None,
        object_filters: Optional[DataSourceFilter] = None,
        sort_by: Optional[str] = None,
        session: Optional[OperableSession] = None,
        **kwargs
    ) -> tuple[Iterable[DataObject], int]:
        if page_size is None:
            page_size = self.get_page_size()
        filter_string = self.__get_filter_string(object_filters)
        sort_string = self.__get_sort_string(sort_by)
        issues, total = self.__client_factory().get_list_page(
            object_type,
            page=page_number,
            page_size=page_size,
            filter_string=f'{filter_string} {sort_string}'.strip()
        )
        converted_issues, _ = self.__jc_factory().convert_list(issues)
        return converted_issues, total

    def get_list(
        self,
        object_type: str,
        object_filters: Optional[DataSourceFilter] = None,
        session: Optional[OperableSession] = None,
        **kwargs
    ) -> Iterable[DataObject]:

        page = 1
        page_size = self.get_page_size()
        client = self.__client_factory()
        jc_converter = self.__jc_factory()
        filter_string = self.__get_filter_string(object_filters)

        while True:
            issues, _ = client.get_list_page(
                object_type,
                page,
                page_size,
                filter_string=filter_string + ' ' + self.__get_sort_string(None)
            )
            (results_page, _) = jc_converter.convert_list(issues)

            yield from results_page
            if len(results_page) < page_size:
                break

            page += 1

    @property
    @cache
    def relationship_config(self) -> dict[str, RelationshipConfig]:
        field_mappings = self.get_fields()
        return {
            'issue': RelationshipConfig(
                to_one={
                    field['system_name']: field['relation']
                    for _, field in field_mappings.items()
                    if field['relation'] is not None
                }
            )
        }

    def get_to_one_relation(
        self,
        source: DataObject,
        relationship_name: str
    ) -> Optional[DataObject]:
        # If we are here then the relationship has not been initialised
        return None

    def get_to_many_relations(
        self,
        source: DataObject,
        relationship_name: str
    ) -> Iterable[DataObject]:
        return []



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import numpy as np

import pandas as pd

from tol.core import OperableDataSource


def convert_excel_to_json(file, sheet_name):
    excel_data = pd.read_excel(file, sheet_name=sheet_name)
    excel_data.replace({np.nan: None}, inplace=True)
    return excel_data.to_dict(orient='records')


def convert_excel_to_valid_json_string(file, sheet_name) -> str:
    """Converts all (date)times to strings"""

    excel_data = pd.read_excel(file, sheet_name=sheet_name)
    excel_data.replace({np.nan: None}, inplace=True)
    return excel_data.to_json(orient='records', date_format='iso')


def __key_is_datetime(
    host: OperableDataSource,
    object_type: str,
    key: str
) -> bool:

    if '.' not in key:
        attr_types = host.attribute_types.get(object_type, {})
        field_type = attr_types.get(key)

        return field_type is not None and 'date' in field_type

    relation, tail = key.split('.', maxsplit=1)
    r_config = host.relationship_config[object_type]
    r_type = r_config.to_one[relation]

    return __key_is_datetime(host, r_type, tail)


def __get_datetime_keys(
    host: OperableDataSource,
    object_type: str,
    keys: list[str]
) -> list[str]:
    """
    the list of column names, out of the selected,
    that may contain timezone info
    """

    return [
        f for f in keys
        if __key_is_datetime(
            host,
            object_type,
            f
        )
    ]


def __make_tz_unaware(
    df: pd.DataFrame,
    datetime_columns: list[str]
) -> pd.DataFrame:
    """
    makes all `datetime` values timezone-unaware in
    a pandas `DataFrame`.
    """

    for column in datetime_columns:
        # (copy and) convert every cell to datetime, with NA if failed
        coerced = pd.to_datetime(df[column], errors='coerce', utc=True)
        # get the 'bitmask' of cells that are datetimes
        datetime_mask = coerced.notna()
        # apply `tz_convert` using the bitmask
        df.loc[datetime_mask, column] = coerced[datetime_mask].dt.tz_convert(None)

    return df



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os
from tempfile import NamedTemporaryFile
from typing import Any, Protocol

from minio import Minio

from .excel_datasource import ExcelDataSource


class S3Fetcher(Protocol):
    def __call__(
        self,
        s3_filename: str,
        s3_bucket: str,
        local_filename: str,
    ) -> None:
        ...


def fetch_from_s3(
    s3_filename: str,
    s3_bucket: str,
    local_filename: str,
) -> None:

    s3_secure = os.getenv('S3_SECURE', 'true').lower() == 'true'

    minio_client = Minio(
        os.environ['S3_URI'],
        os.environ['S3_ACCESS_KEY'],
        secret_key=os.getenv('S3_SECRET_KEY'),
        secure=s3_secure,
    )

    minio_client.fget_object(s3_bucket, s3_filename, local_filename)


def s3_excel_datasource_factory(
    *,
    s3_filename: str,
    s3_bucket: str,
    sheetname: str,
    s3_fetcher: S3Fetcher = fetch_from_s3,
    **kwargs: Any,
) -> ExcelDataSource:

    with NamedTemporaryFile() as temp_fil:
        filename = temp_fil.name

        s3_fetcher(s3_filename, s3_bucket, filename)

        return ExcelDataSource(
            filename,
            sheetname,
            **kwargs,
        )



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from .labwhere_datasource import LabwhereDataSource  # noqa
from .factory import create_labwhere_datasource  # noqa



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)
from ...core.operator.updater import DataObjectUpdate


class ElasticSampleToElasticSequencingRequestUpdateConverter(
        DataObjectToDataObjectOrUpdateConverter):

    def convert(self, data_object: DataObject) -> Iterable[DataObjectUpdate]:
        specimen = data_object.to_one_relationships['sts_specimen']
        yield (
            None,
            {
                'mlwh_sample': self._data_object_factory(
                    'sample',
                    data_object.id
                ),
                'mlwh_specimen.id': specimen.id
            }
        )



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
# SPDX-License-Identifier: MIT
import re
from typing import Iterable

from benchling_sdk.errors import BenchlingError
from benchling_sdk.models import NamingStrategy

from tol.core import DataObject, DataObjectToDataObjectOrUpdateConverter, DataSourceFilter
from tol.sources.benchling import benchling


class StsSampleToCasmBenchlingConverterFactory:
    STS_OBJECT_MAP = {
        'storage_rack': {
            'identifier': 'id',
            'relationship_identifier': 'storage_rack',
        },
        'sex': {
            'identifier': 'name',
            'relationship_identifier': 'target_species_sex',
        },
        'sampleset': {
            'identifier': 'name',
            'relationship_identifier': 'sampleset',
        },
        'sample_status': {
            'identifier': 'status',
            'relationship_identifier': 'sample_status',
        },
        'hazard_group': {
            'identifier': 'level',
            'relationship_identifier': 'hazard_group',
        },
        'target_species': {
            'identifier': 'scientific_name',
            'relationship_identifier': 'target_species',
        },
        'labwhere': {
            'identifier': 'labwhere_id',
            'relationship_identifier': 'storage_rack',
        },
    }
    """
        Map of sts relationship objects to call to.
        These objects could are mainly used to query the
        relationships of the sample for a specific value
    """
    BENCHLING_OBJECT_MAP = {
        'casm_species_v1': {
            'attribute_map': {
                'species_name_v1': 'target_species',
            },
            'primary_attribute': 'species_name_v1',
            'primary_attribute_type': 'schema_field',
            'benchling_relationships': [],
            'sts_relationships': ['target_species'],
            'polymorphic_benchling_relationships': [],
            'converted_value_identifiers': [],
            'stored_values': {},
        },
        'storage': {
            'attribute_map': {
                'barcode': 'labwhere',
            },
            'primary_attribute': 'barcode',
            'primary_attribute_type': 'attribute',
            'benchling_relationships': [],
            'sts_relationships': ['labwhere'],
            'polymorphic_benchling_relationships': [],
            'converted_value_identifiers': [],
            'stored_values': {},
        },
        'casm_compliance_agreement_v1': {
            'attribute_map': {
                'compliance_agreement_id_v1': 'HUMFRE_REFERENCE',
            },
            'primary_attribute': 'compliance_agreement_id_v1',
            'primary_attribute_type': 'schema_field',
            'benchling_relationships': [],
            'sts_relationships': [],
            'polymorphic_benchling_relationships': [],
            'converted_value_identifiers': [],
            'stored_values': {},
        },
        'casm_user_v1': {
            'attribute_map': {
                'email_username_v1': 'SANGER_RESPONSIBLE_SCIENTIST',
            },
            'primary_attribute': 'email_username_v1',
            'primary_attribute_type': 'schema_field',
            'benchling_relationships': [],
            'sts_relationships': [],
            'polymorphic_benchling_relationships': [],
            'converted_value_identifiers': [],
            'stored_values': {},
        },
        'casm_donor_v1': {
            'attribute_map': {
                'id_donor_casm_v1': 'ID_DONOR_CASM',
                'species_v1': 'casm_species_v1',
                'sex_v1': 'sex',
            },
            'primary_attribute': 'id_donor_casm_v1',
            'primary_attribute_type': 'schema_field',
            'benchling_relationships': ['casm_species_v1'],
            'sts_relationships': ['sex'],
            'polymorphic_benchling_relationships': [],
            'converted_value_identifiers': [],
            'stored_values': {},
            'naming_strategy': NamingStrategy.REPLACE_NAMES_FROM_PARTS
        },
        'casm_tissue_v1': {
            'attribute_map': {
                'donor_id_v1': 'casm_donor_v1',
                'tissue_type_v1': 'TISSUE_PHENOTYPE',
                'age_v1': 'SPECIMEN_AGE_YEARS',
                'foetal_tissue_v1': 'FETAL_TISSUE',
                'disease_status_v1': 'WILDTYPE_DISEASE',
                'cancer_type_v1': 'TISSUE_HISTOLOGY',
                'id_tissue_casm_v1': 'ID_TISSUE_CASM',
                'country_of_origin_v1': 'COUNTRY_OF_ORIGIN',
            },
            'primary_attribute': 'id_tissue_casm_v1',
            'primary_attribute_type': 'schema_field',
            'benchling_relationships': ['casm_donor_v1'],
            'sts_relationships': [],
            'polymorphic_benchling_relationships': [],
            'converted_value_identifiers': [],
            'stored_values': {},
            'naming_strategy': NamingStrategy.REPLACE_NAMES_FROM_PARTS
        },
        'casm_sample_metadata_v1': {
            'attribute_map': {
                'tissue_id_v1': 'casm_tissue_v1',
                'compliance_agreement_v1': 'casm_compliance_agreement_v1',
                'responsible_scientist_v1': 'casm_user_v1',
                'tissue_preparation_v1': 'TISSUE_PREPARATION',
                'sts_id_v1': 'id',
                'collaborator_name_v1': 'COLLABORATOR_NAME',
                'responsible_pi_v1': 'SANGER_RESPONSIBLE_PI',
                'sample_set_id_v1': 'sampleset'
            },
            'primary_attribute': 'sts_id_v1',
            'primary_attribute_type': 'schema_field',
            'benchling_relationships': [
                'casm_tissue_v1',
                'casm_compliance_agreement_v1',
                'casm_user_v1'
            ],
            'sts_relationships': ['sampleset'],
            'polymorphic_benchling_relationships': [],
            'converted_value_identifiers': [],
            'stored_values': {},
            'naming_strategy': NamingStrategy.REPLACE_NAMES_FROM_PARTS
        },
        'casm_sample_v1': {
            'attribute_map': {
                'sample_metadata_id_v1': 'casm_sample_metadata_v1',
                'sample_type_v1': 'sample_format',
                'date_created_v1': 'created_on',
                'hazard_group_v1': 'hazard_group',
                'genetically_modified_v1': 'genetically_modified',
                'status_manual_v1': 'sample_status',
                'programme_id_manual_v1': 'INTERNAL_CASM_SAMPLE_NAME',
                'id_sample_casm_manual_v1': 'ID_SAMPLE_CASM'
            },
            'primary_attribute': 'id_sample_casm_manual_v1',
            'primary_attribute_type': 'schema_field',
            'benchling_relationships': [
                'casm_sample_metadata_v1',
            ],
            'sts_relationships': [
                'sample_status',
                'hazard_group',
            ],
            'polymorphic_benchling_relationships': [],
            'converted_value_identifiers': [],
            'stored_values': {},
            'naming_strategy': NamingStrategy.REPLACE_NAMES_FROM_PARTS
        },
        'casm_programme_id_v1': {
            'attribute_map': {
                'sample_id_v1': 'casm_sample_v1',
                'programme_id_v1': 'INTERNAL_CASM_SAMPLE_NAME',
                'id_sample_casm_v1': 'ID_SAMPLE_CASM'
            },
            'primary_attribute': 'sample_id_v1',
            'primary_attribute_type': 'schema_field',
            'benchling_relationships': [
                'casm_sample_v1',
            ],
            'sts_relationships': [],
            'polymorphic_benchling_relationships': [],
            'converted_value_identifiers': [],
            'stored_values': {},
        },
        'casm_sample_status_v1': {
            'attribute_map': {
                'sample_id_v1': 'casm_sample_v1',
                'status_v1': 'sample_status'
            },
            'primary_attribute': 'sample_id_v1',
            'primary_attribute_type': 'schema_field',
            'benchling_relationships': [
                'casm_sample_v1',
            ],
            'sts_relationships': [
                'sample_status'
            ],
            'polymorphic_benchling_relationships': [],
            'converted_value_identifiers': [],
            'stored_values': {},
        },
        '12x12_box': {
            'attribute_map': {
                'barcode': 'storage_rack',
                'parent_storage_id': 'storage'
            },
            'primary_attribute': 'barcode',
            'primary_attribute_type': 'attribute',
            'benchling_relationships': [
                'storage',
            ],
            'sts_relationships': [
                'storage_rack'
            ],
            'polymorphic_benchling_relationships': [],
            'converted_value_identifiers': [],
            'stored_values': {},
        },
        'casm_96_well_plate_v1': {
            'attribute_map': {
                'barcode': 'storage_rack',
                'parent_storage_id': 'storage'
            },
            'primary_attribute': 'barcode',
            'primary_attribute_type': 'attribute',
            'benchling_relationships': [
                'storage',
            ],
            'sts_relationships': [
                'storage_rack'
            ],
            'polymorphic_benchling_relationships': [],
            'converted_value_identifiers': [],
            'stored_values': {},
        },
        'casm_tube_v1': {
            'attribute_map': {
                'barcode': 'tubeid',
                'parent_storage_id': 'box_and_position'
            },
            'primary_attribute': 'barcode',
            'primary_attribute_type': 'schema_field',
            'benchling_relationships': [
                '12x12_box',
            ],
            'sts_relationships': [],
            'polymorphic_benchling_relationships': [],
            'converted_value_identifiers': [],
            'concatenated_values': ['box_and_position'],
            'stored_values': {},
        },
        'casm_well_v1': {
            'attribute_map': {
                'barcode': 'plate_and_location_non_relationship',
                'parent_storage_id': 'plate_and_location'
            },
            'primary_attribute': 'barcode',
            'primary_attribute_type': 'schema_field',
            'benchling_relationships': [
                'casm_96_well_plate_v1',
            ],
            'sts_relationships': ['storage_rack'],
            'polymorphic_benchling_relationships': [],
            'converted_value_identifiers': [],
            'concatenated_values': ['plate_and_location_non_relationship', 'plate_and_location'],
            'stored_values': {},
        },
        'transfer': {
            'attribute_map': {
                'source_entity_id': 'casm_sample_v1',
                'destination_container_id': 'container',
                'transfer_quantity': 'VOLUME_UL',
                'transfer_concentration': 'CONCENTRATION_NG_UL',
            },
            'primary_attribute': None,
            'benchling_relationships': ['casm_sample_v1'],
            'sts_relationships': [],
            'polymorphic_benchling_relationships': [
                'container'
            ],
            'converted_value_identifiers': [],
            'concatenated_values': [],
            'stored_values': {},
        }
    }

    """
     Map of benchling objects to transform based on sts attributes.
     If only stored_values are present then the object is mainly used for storing results in memory
    """

    CONCATENATED_VALUES = {
        'plate_and_location': {
            'values': [
                'casm_96_well_plate_v1',
                'TUBE_WELL_POSITION'
            ],
            'separator': ':'
        },
        'plate_and_location_non_relationship': {
            'values': [
                'storage_rack',
                'TUBE_WELL_POSITION'
            ],
            'separator': ':'
        },
        'box_and_position': {
            'values': [
                '12x12_box',
                'TUBE_WELL_POSITION'
            ],
            'separator': ':'
        }
    }
    """
    Map of the values that need to be concatenated for the use within benchling
    """

    VALUE_REPLACEMENTS = {
        'sex_v1': {
            'MALE': 'Male',
            'FEMALE': 'Female',
            'NOT_PROVIDED': 'Unknown'
        },
        'responsible_pi_v1': {
            'default': 'other',
            'da1': 'David Adams',
            'im3': 'Inigo Martincorena',
            'jn5': 'Jyoti Nangalia',
            'ly2': 'Lucy Yates',
            'mg12': 'Mathew Garnett',
            'mrs': 'Mike Stratton',
            'pc8': 'Peter Campbell',
            'pj3': 'Phil Jones',
            'rr11': 'Raheleh Rahbari',
            'sb31': 'Sam Behjati',
            'tjm': 'Thomas Mitchell',
        },
        'genetically_modified_v1': {
            'default': 'No'
        },
        'status_manual_v1': {
            'ACCEPTED': 'Available'
        },
        'status_v1': {
            'ACCEPTED': 'Available'
        },
        'species_name_v1': {
            'Canis lupus familiaris': 'Canis familiaris'
        },
        'sample_type_v1': {
            'inactivated biological sample from infectious organism': 'Tissue',
            'live biological sample from infectious organism': 'Tissue',
            'biological sample / tissue from non-infectious organism': 'Tissue',
            'default': 'DNA'
        }
    }
    """
        Map of replacements for string objects. Mainly used for data cleanup
    """

    DESTINATION_OBJECT_TYPES = {
        'box_or_plate': {
            'RACK_TUBE': '12x12_box',
            'PLATE_WELL': 'casm_96_well_plate_v1'
        },
        'container': {
            'RACK_TUBE': 'casm_tube_v1',
            'PLATE_WELL': 'casm_well_v1'
        }
    }
    """
        Map of the dynamic object types
    """

    POLYMORPHIC_RELATIONSHIP_OBJECT_TYPES = {
        'container': {
            'RACK_TUBE': 'casm_tube_v1',
            'PLATE_WELL': 'casm_well_v1'
        }
    }
    """
        Map of the polymorphic relationship objects
    """

    destination_object_type: str
    fields: Iterable[any]

    def __init__(
            self,
            destination_object_type: str = '',
            previous_object_type: str = '',
            previous_objects: list = None,
            detect_destination: bool = False,
            detect_destination_type: str = '',
    ):
        self.detect_destination = detect_destination
        self.benchling = benchling()
        if not detect_destination:
            self.populate_destination(destination_object_type)
        elif detect_destination_type == '':
            raise Exception(
                'Configuration error of flow: Please include a '
                'detect_destination_type to auto-detect'
            )
        else:
            self.detect_destination_type = detect_destination_type

        if previous_object_type and previous_objects:
            for previous_object in previous_objects:
                object_map = self.BENCHLING_OBJECT_MAP[previous_object_type]
                identifier = object_map['primary_attribute']
                key = getattr(previous_object, identifier)
                object_map['stored_values'][key] = previous_object

    def populate_destination(self, destination_object_type):
        self.destination_object_type = destination_object_type

        if 'transfer' != self.destination_object_type:
            benchling_type = self.benchling.benchling_types[self.destination_object_type]
            self.fields = self.benchling.schemas[benchling_type][self.destination_object_type]

    def get_converter_class(self) -> DataObjectToDataObjectOrUpdateConverter:
        factory = self

        class StsSampleToCasmBenchlingConverter(DataObjectToDataObjectOrUpdateConverter):
            def convert(self, data_object: DataObject) -> Iterable[DataObject]:
                sample = data_object

                if factory.detect_destination:
                    destination_object_type = self._get_destination_object_type(
                        sample=sample,
                        detect_destination_type=factory.detect_destination_type
                    )
                    factory.populate_destination(destination_object_type)

                object_map = factory.BENCHLING_OBJECT_MAP[factory.destination_object_type]
                if not self._does_object_exist(
                    factory.destination_object_type,
                    sample,
                    object_map
                ):
                    self._populate_relationships(sample, object_map)

                    object_attributes = self._populate_object_attributes(object_map, sample)

                    if 'transfer' != factory.destination_object_type:
                        primary_attribute = self._get_object_primary_attribute_value(
                            object_map,
                            sample
                        )
                        object_map['converted_value_identifiers'] = \
                            object_map['converted_value_identifiers'] + [primary_attribute]

                    if 'naming_strategy' in object_map and object_map['naming_strategy']:
                        object_attributes['naming_strategy'] = object_map['naming_strategy']

                    yield self._data_object_factory(
                        factory.destination_object_type,
                        sample.id,
                        attributes=object_attributes
                    )

            @staticmethod
            def _get_destination_object_type(
                    sample,
                    detect_destination_type: str,
                    raise_exception: bool = True
            ) -> str | None:
                """
                Determines the destination object type for a sample.

                Args:
                    sample - the sample used to determine the destination object type
                    detect_destination_type - the type of destination to detect
                    raise_exception - whether to raise an exception if the
                    destination type is unsupported

                Expects:
                    Exception - if the sample has an unsupported destination type
                    and raise_exception is True

                Returns:
                    str | None - the detected destination object type or None if not found
                    and raise_exception is False
                """
                if (
                        hasattr(sample, 'manifest')
                        and hasattr(sample.manifest, 'manifest_type')
                        and sample.manifest.manifest_type in factory.
                        DESTINATION_OBJECT_TYPES[detect_destination_type]
                ):
                    return factory.DESTINATION_OBJECT_TYPES[detect_destination_type][
                        sample.manifest.manifest_type]

                if raise_exception:
                    raise Exception(
                        f'Sample is not ready for import: Sample #{sample.id}'
                        f' has unsupported destination type for dynamic conversion'
                    )

                return None

            @staticmethod
            def _sanitize_attribute(key: str, value: any, object_type_override: str = ''):
                """
                    This static method sanitizes an attribute making sure it's the
                    correct type expected by Benchling, it will also transform the value of
                    the attribute to a predetermined safe value for
                    Benchling this is configured in VALUE_REPLACEMENTS.

                    Args:
                         key -  This argument specifies the key of the attribute,
                         value -  This argument specifies the value of the attribute to be cleaned
                         object_type_override – This argument specifies the cleanup actions to
                         be performed if the attribute does not belong to
                         the destination object of the converter.

                    Return:
                        Any - depends on the value provided and the cleanup performed
                """
                fields = getattr(factory, 'fields', [])

                if '' != object_type_override:
                    benchling_type = factory.benchling.benchling_types[object_type_override]
                    if (
                            benchling_type
                            and object_type_override in factory.benchling.schemas[benchling_type]
                    ):
                        fields = factory.benchling.schemas[benchling_type][object_type_override]

                if fields and key in fields:
                    if 'int' == fields[key]['type']:
                        if value:
                            value = int(value)
                        else:
                            value = 0

                    if 'str' == fields[key]['type']:
                        if value:
                            value = str(value)

                        if key in factory.VALUE_REPLACEMENTS:
                            if value in factory.VALUE_REPLACEMENTS[key]:
                                value = factory.VALUE_REPLACEMENTS[key][value]
                            elif 'default' in factory.VALUE_REPLACEMENTS[key]:
                                value = factory.VALUE_REPLACEMENTS[key]['default']

                        if fields[key]['is_multi']:
                            value = [value]

                    if 'genetically_modified' == key:
                        value = factory.VALUE_REPLACEMENTS[key]['default']

                return value

            @staticmethod
            def _get_sts_relationship_attribute_value(relationship_object_identifier: str, sample):
                """
                Retrieves the attribute value for an STS relationship.

                Args:
                    relationship_object_identifier - the identifier of the relationship object
                    sample - the sample from which to retrieve the relationship attribute value

                Returns:
                    str | None - the attribute value of the relationship object, or None
                """
                attribute_value = None
                relationship_object_map = factory.STS_OBJECT_MAP[relationship_object_identifier]
                sts_relationship = getattr(
                    sample,
                    relationship_object_map['relationship_identifier'],
                    None
                )

                if (
                    isinstance(sts_relationship, Iterable)
                    and not isinstance(sts_relationship, str)
                ):
                    relationship_object = next(iter(sts_relationship), None)
                else:
                    relationship_object = sts_relationship

                if relationship_object is not None:
                    attribute_value = getattr(
                        relationship_object,
                        relationship_object_map['identifier'],
                        None
                    )

                return attribute_value

            def _populate_concatenated_attributes(self, sample, object_map):
                """
                Populates concatenated attributes for a given sample.

                Args:
                    sample - the sample whose concatenated attributes need to be populated
                    object_map - a mapping of attributes relevant to the sample

                Returns:
                    None
                """

                self._populate_relationships(sample, object_map)

                for key, attribute_mapping in object_map['attribute_map'].items():
                    if (
                        attribute_mapping in factory.CONCATENATED_VALUES
                        and sample.attributes.get(attribute_mapping, None) is None
                    ):
                        separator = factory.CONCATENATED_VALUES[attribute_mapping]['separator']

                        # Strip out any trailing 0 for the TUBE_WELL_POSITION as benchling strips
                        # this out on save so it breaks any search queries for bar codes
                        values = [
                            re.sub(
                                r'([A-Za-z]+)0',
                                r'\1',
                                sample.attributes.get(attribute, '') or ''
                            ) if attribute == 'TUBE_WELL_POSITION' else sample.attributes.get(
                                attribute, ''
                            ) for attribute in factory.CONCATENATED_VALUES[
                                attribute_mapping]['values']
                        ]

                        sample.attributes[attribute_mapping] = separator.join(filter(None, values))

            def _populate_sts_relationships(self, sample, object_map):
                """
                    This method populates the attributes with the values from the sts_relationships

                    Args:
                        sample: sample data object from sts

                    Returns:
                        None

                    Expects a StopIteration exception if the relationship does not have a value
                """
                for relationship_object_identifier in object_map['sts_relationships']:
                    if sample.attributes.get(relationship_object_identifier, None) is None:
                        sample.attributes[relationship_object_identifier] = \
                            self._get_sts_relationship_attribute_value(
                                relationship_object_identifier, sample)

            def _does_object_exist(self, destination_object_type, sample, object_map):
                """
                Checks if the object all ready exists within the
                Benchling ecosystem or is already loaded into memory

                Args:
                    destination_object_type - type of object we are looking for
                    sample - the sample we are using to get the search values

                Expects:
                    StopIteration - if no object is returned from Benchling
                """

                stored_values = object_map['stored_values']
                converted_value_ids = object_map['converted_value_identifiers']

                if 'transfer' == destination_object_type:
                    return self._check_sample_transfers_done(sample, object_map)
                else:
                    attribute = self._get_object_primary_attribute_value(object_map, sample)

                    if attribute in stored_values or attribute in converted_value_ids:
                        return True

                    benchling_object_id = self._get_benchling_object_id(
                        object_type=destination_object_type,
                        search_identifier=object_map['primary_attribute'],
                        search_value=attribute,
                        add_to_return=True
                    )

                    if benchling_object_id is not None:
                        factory.BENCHLING_OBJECT_MAP[factory.destination_object_type][
                            'stored_values'][attribute] = benchling_object_id

                        return True

                return False

            def _check_sample_transfers_done(self, sample, object_map) -> bool:
                """
                Checks if a sample has any transfers by retrieving the container for the sample.

                Args:
                    sample - the sample to check for completed transfers
                    object_map - a mapping of attributes relevant to the sample

                Expects:
                    BenchlingError - if an error occurs while retrieving container contents

                Returns:
                    bool - True if the container has contents, False otherwise

                Raises:
                    Exception: If teh sample has no container in Bechnling
                """
                self._populate_relationships(sample, object_map)

                container_id = sample.attributes.get(
                    object_map['attribute_map']['destination_container_id']
                )
                if not container_id:
                    raise Exception(
                        f'Sample: {sample.id} not ready for transfer as '
                        f'it does not have a container registered in benchling'
                    )

                contents_found = True
                try:
                    contents = factory.benchling.get_container_contents(container_id)

                    if not contents:
                        contents_found = False
                except BenchlingError:
                    contents_found = False

                return contents_found

            def _get_object_primary_attribute_value(self, object_map, sample):
                """
                Retrieves the primary attribute value for a given object map from the sts sample.

                Args:
                    object_map - a mapping of attributes relevant to the object
                    sample - the sample from which to retrieve the primary attribute value

                Returns:
                    str | None - the primary attribute value, or None if not found
                """
                benchling_attribute_identifier = object_map['primary_attribute']
                sts_attribute_identifier = object_map['attribute_map'][
                    benchling_attribute_identifier]

                if (
                        sts_attribute_identifier in object_map['sts_relationships']
                        and sts_attribute_identifier in factory.STS_OBJECT_MAP
                ):
                    self._populate_sts_relationships(sample, object_map)
                elif (
                    'concatenated_values' in object_map
                    and sts_attribute_identifier in object_map['concatenated_values']
                ):
                    self._populate_concatenated_attributes(sample, object_map)

                if sts_attribute_identifier in ['id', 'sts_id']:
                    attribute_value = sample.id
                else:
                    attribute_value = sample.attributes.get(sts_attribute_identifier, None)

                attribute_value = self._sanitize_attribute(
                    sts_attribute_identifier,
                    attribute_value
                )

                return attribute_value

            def _populate_relationships(self, sample, object_map):
                """
                Populates the relationships for a given sample based on the object_map

                Args:
                    sample - the sample whose relationships need to be populated
                    object_map - a mapping of attributes relevant to the sample

                Returns:
                    None
                """
                self._populate_benchling_relationships(sample, object_map)
                self._populate_sts_relationships(sample, object_map)

            def _populate_benchling_relationships(self, sample, object_map):
                """
                Populates the STS sample object attributes with Benchling-compatible values.
                This ensures Benchling relationship IDs replace human-readable elements.

                Args:
                    sample: Sample data object from STS.
                    object_map: The nested JSON config from the factory.

                Raises:
                    Exception: If a required relationship is missing or cannot be populated.

                Returns:
                    None
                """
                self._populate_polymorphic_benchling_relationships(sample, object_map)

                for benchling_object_identifier in object_map['benchling_relationships']:
                    relationship_object_map = factory.BENCHLING_OBJECT_MAP[
                        benchling_object_identifier]
                    if sample.attributes.get(benchling_object_identifier, None) is None:
                        search_value = self._get_object_primary_attribute_value(
                            relationship_object_map,
                            sample
                        )

                        if search_value is not None:
                            if search_value in relationship_object_map['stored_values']:
                                benchling_object_id = relationship_object_map['stored_values'][
                                    search_value]
                            else:
                                benchling_object_id = self._get_benchling_object_id(
                                    object_type=benchling_object_identifier,
                                    search_identifier=relationship_object_map['primary_attribute'],
                                    search_value=search_value
                                )

                            if benchling_object_id is not None:
                                sample.attributes[benchling_object_identifier] = \
                                    benchling_object_id
                                continue

                        raise Exception(
                            f'Sample not ready for import: {sample.id} is missing the '
                            f'benchling relationship for {benchling_object_identifier}'
                        )

            def _populate_object_attributes(self, object_map, sample):
                """
                Populates the attributes for an object using the given sample.

                Args:
                    object_map - a mapping of attributes relevant to the object
                    sample - the sample from which to populate the object attributes

                Returns:
                    dict - a dictionary of populated object attributes
                """
                attribute_map = object_map['attribute_map']
                self._populate_concatenated_attributes(sample, object_map)

                object_attributes = {
                    key: (
                        sample.id
                        if 'id' == attr_mapping
                        else sample.attributes.get(attr_mapping)
                    )
                    for key, attr_mapping in attribute_map.items()
                }
                self._sanitize_attributes(object_attributes)
                return object_attributes

            def _populate_polymorphic_benchling_relationships(self, sample, object_map):
                """
                Populates the 'benchling_relationships' list in object_map with detected
                polymorphic relationships for the given sample.

                Args:
                    sample: The sample object being processed.
                    object_map (dict): A mapping containing 'polymorphic_benchling_relationships'
                                       and 'benchling_relationships' lists.

                The function iterates over 'polymorphic_benchling_relationships', determines
                the destination object type, and appends it to 'benchling_relationships'
                if it's not already present.
                """
                for relationship in object_map.get('polymorphic_benchling_relationships', []):
                    relationship_object_type = self._get_destination_object_type(
                        sample=sample,
                        detect_destination_type=relationship,
                        raise_exception=False
                    )

                    if (
                        relationship_object_type
                        and relationship_object_type not in object_map['benchling_relationships']
                    ):

                        key_for_relationship_object_type = next(
                            (
                                key for key, value in object_map['attribute_map'].items()
                                if relationship == value
                            ),
                            None
                        )

                        if key_for_relationship_object_type:
                            object_map['benchling_relationships'].append(relationship_object_type)
                            object_map['attribute_map'][key_for_relationship_object_type] = \
                                relationship_object_type

            def _get_benchling_object_id(
                self,
                object_type: str,
                search_identifier: str,
                search_value: str,
                add_to_return: bool = False
            ) -> str | None:
                """
                This method is used to get the benchling object id for its given args

                Args:
                    object_type: String identifying the benchling object type
                    search_identifier: The identifier of the attribute of the benchling object
                    search_value: The values we are searching for
                """
                filter_object = DataSourceFilter()
                if 'custom_entity' == factory.benchling.benchling_types[object_type]:
                    schema_filter = DataSourceFilter()
                    schema_filter.and_ = {search_identifier: {'eq': {'value': search_value}}}
                    filter_object.and_ = {'schema_fields': schema_filter}
                elif (factory.benchling.benchling_types[object_type]
                      in ['box', 'plate', 'container', 'location']):
                    if 'barcode' == search_identifier:
                        search_identifier = 'barcodes'
                    filter_object.and_ = {
                        search_identifier: {'in_list': {'value': [search_value]}}
                    }
                elif factory.benchling.benchling_types[object_type] in ['assay_result']:
                    filter_object.and_ = {'entity_id': {'eq': {'value': [search_value]}}}
                else:
                    raise Exception(
                        f'Configuration error: Unsupported search of type {object_type}'
                    )

                try:
                    benchling_object = next(
                        iter(
                            factory.benchling.get_list(
                                object_type,
                                filter_object
                            )
                        )
                    )

                    if add_to_return:
                        self._return_objects.append(benchling_object)

                    return benchling_object.id
                except StopIteration:
                    return None

            def _sanitize_attributes(self, object_attributes):
                for key, value in object_attributes.items():
                    object_attributes[key] = self._sanitize_attribute(key, value)

        return StsSampleToCasmBenchlingConverter



# SPDX-FileCopyrightText: 2026 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import json
import os

from tol.core import (
    core_data_object
)
from tol.google_sheets import (
    GoogleSheetDataSource
)


def library_batch_ids(**kwargs) -> GoogleSheetDataSource:
    gsds = GoogleSheetDataSource({
        'client_secrets': json.loads(os.getenv('GOOGLE_CLIENT_SECRETS')),
        'sheet_key': '1IyGPD9dF51kUL2mgW6u0r4CNPy7lsTGC2UfD_M58lNo',
        'mappings': {
            'sequencing_request': {
                'worksheet_name': 'Sheet1',
                'columns': {
                    'id': {
                        'heading': 'sequencing_request_id',
                        'type': 'str',
                    },
                    'library_batch_id': {
                        'heading': 'library_batch_id',
                        'type': 'str',
                    },
                },
                'header_row': 1,
                'data_start_row': 2,
            }
        }
    })
    core_data_object(gsds)
    return gsds



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Any, Iterable, Optional

from ._filterable import _Filterable

if typing.TYPE_CHECKING:
    from ..datasource_filter import DataSourceFilter
    from ..session import OperableSession


class GroupStatter(_Filterable, ABC):
    @abstractmethod
    def get_group_stats(
        self,
        object_type: str,
        group_by: list[str],
        stats_fields: list[str] = [],
        stats: list[str] = ['min', 'max'],
        object_filters: Optional[DataSourceFilter] = None,
        session: Optional[OperableSession] = None
    ) -> Iterable[dict[Any, int]]:
        """
        Gets stats for results that are matched by the (optional) filter,
        broken down by the group_by parameter
        """



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections import ChainMap
from pathlib import Path
from typing import Any, Iterable

from yaml import safe_load

from .model import (
    AttributeConfig,
    DestinationConfig,
    YamlConfig,
)
from ...core import (
    DataObject,
    DataObjectFactory,
    DataObjectToDataObjectOrUpdateConverter,
)


class YamlConverter(DataObjectToDataObjectOrUpdateConverter):
    """
    Used for Tree of Sex.

    Converts `DataObject` instances dynamically, according to
    a YAML specification.
    """

    def __init__(
        self,
        data_object_factory: DataObjectFactory,
        yaml_path: str | Path,
        *,
        destination_object_type: str | None = None,
        pydantic_class: type[YamlConfig] = YamlConfig,
    ) -> None:

        super().__init__(data_object_factory)

        self.__config = self.__load_yaml(
            yaml_path,
            pydantic_class,
        )
        self.__dest_type = destination_object_type

    def convert(
        self,
        input_: DataObject
    ) -> Iterable[DataObject]:

        attributes = self.__convert_attributes(
            input_,
        )
        destination_type = self.__get_destination_type(
            input_,
        )

        yield self._data_object_factory(
            destination_type,
            id_=input_.id,
            attributes=attributes,
        )

    def __get_destination_type(
        self,
        input_: DataObject,
    ) -> str:

        if self.__dest_type:
            return self.__dest_type
        else:
            return input_.type

    def __convert_attributes(
        self,
        input_: DataObject
    ) -> dict[str, Any]:

        attr_dicts = [
            self.__convert_attribute(
                input_,
                attribute_config,
            )
            for attribute_config
            in self.__config.attributes
        ]

        return dict(
            ChainMap(*attr_dicts)
        )

    def __convert_attribute(
        self,
        input_: DataObject,
        attribute_config: AttributeConfig,
    ) -> dict[str, Any]:

        pairs = [
            self.__convert_destination(
                input_,
                d,
                attribute_config.imported_column_name,
            )
            for d in self.__get_destinations(
                attribute_config,
            )
        ]

        return dict(pairs)

    def __get_destinations(
        self,
        attribute_config: AttributeConfig
    ) -> list[DestinationConfig]:

        if isinstance(attribute_config.destination, DestinationConfig):
            return [attribute_config.destination]
        else:
            return attribute_config.destination

    def __convert_destination(
        self,
        input_: DataObject,
        destination_config: DestinationConfig,
        imported_column_name: str,
    ) -> tuple[str, Any]:

        if not self.__should_convert_destination(
            input_,
            imported_column_name,
        ):
            __v = getattr(input_, imported_column_name)

            return destination_config.key, __v

        split_values = self.__get_split_values(
            input_,
            destination_config,
            imported_column_name,
        )

        seperator = destination_config.separator

        converted = seperator.join(
            self.__convert_split_values(
                split_values,
                destination_config,
            )
        )

        return destination_config.key, converted

    def __should_convert_destination(
        self,
        input_: DataObject,
        imported_column_name: str,
    ) -> bool:

        __v = getattr(input_, imported_column_name)

        return (
            __v is not None and isinstance(__v, str)
        )

    def __get_split_values(
        self,
        input_: DataObject,
        destination_config: DestinationConfig,
        imported_column_name: str,
    ) -> list[str]:

        value: str = getattr(
            input_,
            imported_column_name,
        )

        return value.split(
            destination_config.separator,
        )

    def __convert_split_values(
        self,
        values: list[str],
        destination_config: DestinationConfig,
    ) -> list[str]:

        value_map = destination_config.imported_values_map

        return [
            value_map[v] if v in value_map else v
            for v in values
            if self.__include_split_value(
                v,
                destination_config,
            )
        ]

    def __include_split_value(
        self,
        value: str,
        destination_config: DestinationConfig,
    ) -> bool:

        if value in destination_config.ignore:
            return False

        if destination_config.magic_match_all:
            return True

        value_map = destination_config.imported_values_map
        if value in value_map:
            return True

        return self.__allowed_magic_type(
            value,
            destination_config,
        )

    def __allowed_magic_type(
        self,
        value: str,
        destination_config: DestinationConfig,
    ) -> bool:

        for t in destination_config.magic_types:
            try:
                t(value)
                return True
            except ValueError:
                continue

        return False

    def __load_yaml(
        self,
        yaml_path: str | Path,
        pydantic_class: type[YamlConfig],
    ) -> YamlConfig:

        with open(yaml_path, 'r') as yaml_file:
            loaded = safe_load(yaml_file)

        return pydantic_class(**loaded)



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC
from typing import Iterable, List


if typing.TYPE_CHECKING:
    from ..data_object import DataObject


class Enricher(ABC):
    """
    Enriches attributes of one object onto relationships of another object
    """

    @property
    def enriching_fields(self):
        """
        Returns a dictionary of object types and their enriching fields
        e.g. {'species': ['goat_family', 'goat_genome_size']}
        """
        object_enriching_fields = {}
        for object_type, attributes in self.attribute_metadata.items():
            enriching_fields = []
            for attribute, metadata in attributes.items():
                if metadata['available_on_relationships']:
                    enriching_fields.append(attribute)
            if len(enriching_fields) > 0:
                object_enriching_fields[object_type] = enriching_fields
        return object_enriching_fields

    @property
    def relationships_to_enrich(self):
        """
        Returns a dictionary of object types and their relationships to enrich for
        each source object type
        e.g. {'extraction': {'species': ['sts_species', 'benchling_species]}}
        """
        relationships_to_enrich = {}
        for object_type, rc in self.relationship_config.items():
            if rc.to_one is not None:
                for rel_name, related_object_type in rc.to_one.items():
                    if related_object_type not in relationships_to_enrich:
                        relationships_to_enrich[related_object_type] = {}
                    if object_type not in relationships_to_enrich[related_object_type]:
                        relationships_to_enrich[related_object_type][object_type] = []
                    relationships_to_enrich[related_object_type][object_type].append(rel_name)
        return relationships_to_enrich

    def get_enrich_update(
            self,
            enriching_fields: List[str],
            source_data: Iterable[DataObject],
            target_object_type: str
    ):
        """
        Gets the update data for enriching an index with the source fields
        Makes an update if the id of the target attribute is present (candidate_key).

        args:
            source_fields: list of fields to enrich the index with
            source_data: elastic.get_list data returned by the source_index
            target_object_type: the object type of the target index
        """
        for obj in source_data:
            obj_dict = {'id': obj.id}
            # Get the values of the enriching fields
            for field in enriching_fields:
                obj_dict[field] = obj.get_field_by_name(field)
            for target_attribute in self.relationships_to_enrich[obj.type][target_object_type]:
                yield (
                    None,
                    {
                        f'{target_attribute}': obj_dict,
                        f'{target_attribute}.id': obj.id  # The candidate key
                    }
                )

    def enrich(
            self,
            source_object_type: str,
            source_objects: Iterable[DataObject],
            target_object_type: str):
        """
        Enriches the target object type with the enriching fields from the source object type
        """
        if source_object_type in self.enriching_fields:
            updates = self.get_enrich_update(
                self.enriching_fields[source_object_type],
                source_objects,
                target_object_type
            )
            candidate_key_possibilities = [
                f'{rel}.id'
                for rel in self.relationships_to_enrich[source_object_type][target_object_type]
            ]
            self.update(
                target_object_type,
                updates,
                candidate_key_func=lambda x:
                    [next((key for key in candidate_key_possibilities if key in x))]
            )



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
# SPDX-License-Identifier: MIT

import re
from dataclasses import dataclass
from datetime import datetime
from typing import Iterable

from tol.core import DataObject, DataObjectToDataObjectOrUpdateConverter


class IncomingSampleToEnaSampleConverter(DataObjectToDataObjectOrUpdateConverter):

    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        ena_checklist_id: str
        project_name: str

    __slots__ = ['__config']
    __config: Config

    def __init__(self, data_object_factory, config: Config) -> None:
        super().__init__(data_object_factory)
        self.__config = config
        self._data_object_factory = data_object_factory

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        """
        converting the samples DataObject into ENA format
        """
        s = data_object
        attributes = {
            'ENA-CHECKLIST': self.__config.ena_checklist_id,
            'organism part': self.__join_list([
                self.__replace_underscores(v)
                for v in s.attributes.get('ORGANISM_PART', [])
            ]),
            'lifestage': (
                'spore-bearing structure'
                if s.attributes.get('LIFESTAGE') == 'SPORE_BEARING_STRUCTURE'
                else self.__replace_underscores(
                    s.attributes.get('LIFESTAGE'))
            ),
            'project name':
                self.__config.project_name,
            'collected_by': self.__join_list([
                self.__replace_underscores(v)
                for v in s.attributes.get('COLLECTED_BY', [])
            ]),
            'collection date':
                self.__format_date(
                    s.attributes.get('DATE_OF_COLLECTION')),
            'geographic location (country and/or sea)':
                self.__collection_country(s).replace('_', ' '),
            'geographic location (latitude)':
                self.__replace_underscores(
                    str(s.attributes.get('DECIMAL_LATITUDE'))).lower(),
            'geographic location (latitude) units':
                'DD',
            'geographic location (longitude)':
                self.__replace_underscores(
                    str(s.attributes.get('DECIMAL_LONGITUDE'))).lower(),
            'geographic location (longitude) units':
                'DD',
            'geographic location (region and locality)':
                self.__collection_region(s).replace('_', ' '),
            'identified_by': self.__join_list([
                self.__replace_underscores(v)
                for v in s.attributes.get('IDENTIFIED_BY', [])
            ]),
            'habitat':
                self.__replace_underscores(
                    s.attributes.get('HABITAT')),
            'identifier_affiliation': self.__join_list([
                self.__replace_underscores(v)
                for v in s.attributes.get('IDENTIFIER_AFFILIATION', [])
            ]),
            'sex':
                self.__replace_underscores(
                    s.attributes.get('SEX')),
            'relationship':
                self.__replace_underscores(
                    s.attributes.get('RELATIONSHIP')),
            'SYMBIONT':
                'Y' if s.attributes.get('SYMBIONT') == 'SYMBIONT' else 'N',
            'collecting institution': self.__join_list([
                self.__replace_underscores(v)
                for v in s.attributes.get('COLLECTOR_AFFILIATION', [])
            ]),
        }
        if self.__sanitise(s.attributes.get('DEPTH')) != '':
            attributes['geographic location (depth)'] = s.attributes.get('DEPTH')
            attributes['geographic location (depth) units'] = 'm'
        if self.__sanitise(s.attributes.get('ELEVATION')) != '':
            attributes['geographic location (elevation)'] = s.attributes.get('ELEVATION')
            attributes['geographic location (elevation) units'] = 'm'
        if self.__sanitise(s.attributes.get('ORIGINAL_COLLECTION_DATE')) != '':
            attributes['original collection date'] = \
                self.__format_date(s.attributes.get('ORIGINAL_COLLECTION_DATE'))
        if self.__sanitise(s.attributes.get('ORIGINAL_GEOGRAPHIC_LOCATION')) != '':
            attributes['original geographic location'] = \
                self.__replace_underscores(s.attributes.get('ORIGINAL_GEOGRAPHIC_LOCATION'))
        if s.attributes.get('GAL') is not None:
            attributes['GAL'] = s.attributes.get('GAL')
        if s.attributes.get('VOUCHER_ID') is not None:
            attributes['specimen_voucher'] = s.attributes.get('VOUCHER_ID')
        if s.attributes.get('SPECIMEN_ID') is not None:
            attributes['specimen_id'] = s.attributes.get('SPECIMEN_ID')
        if s.attributes.get('GAL_SAMPLE_ID') is not None:
            attributes['GAL_sample_id'] = s.attributes.get('GAL_SAMPLE_ID')
        if s.attributes.get('CULTURE_OR_STRAIN_ID') is not None:
            attributes['culture_or_strain_id'] = s.attributes.get('CULTURE_OR_STRAIN_ID')

        ret = self._data_object_factory(
            data_object.type,
            s.id,
            attributes=attributes,
        )
        yield ret

    def __collection_country(self, data_object: DataObject):
        return re.split(
            r'\s*\|\s*',
            data_object.attributes.get('COLLECTION_LOCATION'))[0]

    def __collection_region(self, data_object: DataObject):
        return ' | '.join(re.split(
            r'\s*\|\s*',
            data_object.attributes.get('COLLECTION_LOCATION'))[1:])

    def __replace_underscores(self, value):
        if type(value) != str:
            return value
        return self.__sanitise(value, '').replace('_', ' ')

    def __sanitise(self, value, default_value=''):
        if value is None:
            return default_value
        return value

    def __join_list(self, value_list):
        if value_list is None:
            return ''
        if not isinstance(value_list, list):
            return str(value_list)
        return ' | '.join(str(v) for v in value_list)

    def __format_date(self, value):
        """Format date to YYYY-mm-dd format"""
        if value is None:
            return ''
        if isinstance(value, datetime):
            return value.strftime('%Y-%m-%d')

        return str(value)



# SPDX-FileCopyrightText: 2021 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT
import os

REDPANDA_URL = os.getenv('REDPANDA_URL', '127.0.0.1')
REDPANDA_API_KEY = os.getenv('REDPANDA_API_KEY', 'test')
REDPANDA_CREATE_LABWARE_SUBJECT = 'create-labware'

RABBITMQ_HOST = os.environ.get('RABBITMQ_HOST', '127.0.0.1')
RABBITMQ_PORT = os.getenv('RABBITMQ_PORT', '5671')
RABBITMQ_USERNAME = os.getenv('RABBITMQ_USERNAME', 'psd')
RABBITMQ_PASSWORD = os.getenv('RABBITMQ_PASSWORD', 'psd')
RABBITMQ_VHOST = os.getenv('RABBITMQ_VHOST', 'tol')
RABBITMQ_EXCHANGE = os.getenv('RABBITMQ_EXCHANGE', 'tol-team.tol')
RABBITMQ_ROUTING_KEY = os.getenv('RABBITMQ_ROUTING_KEY', 'crud.1')
RABBITMQ_USE_SSL = (os.getenv('RABBITMQ_USE_SSL', 'False').lower() == 'true')
RABBITMQ_PUBLISH_RETRY_DELAY = int(os.getenv('RABBITMQ_PUBLISH_RETRY_DELAY', 5))
RABBITMQ_PUBLISH_RETRIES = int(os.getenv('RABBITMQ_PUBLISH_RETRIES', 36))

TOL_FEEDBACK_QUEUE = os.getenv('TOL_FEEDBACK_QUEUE', 'tol.feedback')

CREATE_LABWARE_MESSAGE_SCHEMA_VERSION = os.getenv('CREATE_LABWARE_MESSAGE_SCHEMA_VERSION',
                                                  'latest')
UPDATE_LABWARE_MESSAGE_SCHEMA_VERSION = os.getenv('UPDATE_LABWARE_MESSAGE_SCHEMA_VERSION',
                                                  'latest')



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import json
import os

from ..core import (
    core_data_object
)
from ..google_sheets import (
    GoogleSheetDataSource
)


def informatics(**kwargs) -> GoogleSheetDataSource:
    gsds = GoogleSheetDataSource({
        'client_secrets': json.loads(os.getenv('GOOGLE_CLIENT_SECRETS')),
        'sheet_key': '1RKubj10g13INd4W7alHkwcSVX_0CRvNq0-SRe21m-GM',
        'mappings': {
            'tolid': {
                'worksheet_name': 'Status',
                'columns': {
                    'id': {
                        'heading': 'sample',
                        'type': 'str'
                    },
                    'status_summary': {
                        'heading': 'statussummary',
                        'type': 'str'
                    },
                    'status': {
                        'heading': 'status',
                        'type': 'str'
                    },
                    'gscope_coverage': {
                        'heading': 'long read cov (gscope)',
                        'type': 'float'
                    },
                    'estimated_genome_size': {
                        'heading': 'est. size (Mb)',
                        'type': 'float'
                    }

                },
                'header_row': 1,
                'data_start_row': 2
            }
        }
    })
    core_data_object(gsds)
    return gsds



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass

from tol.core import Validator
from tol.core.data_object import DataObject


class UniqueValueCheckValidator(Validator):
    """
    Validates an incoming stream of `DataObject` instances.
    For each data object (sample) it checks if the GAL column has only one value
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        field: str

    __slots__ = ['__config', '_cached']
    __config: Config
    _cached: str | None

    def __init__(self, config: Config, **kwargs) -> None:
        super().__init__()
        self.__config = config
        self._cached = None

    def _validate_data_object(self, obj: DataObject) -> None:
        # This function is used to check if the data object has a single value for GAL column

        if not self._cached:
            self._cached = obj.attributes.get(self.__config.field)

        else:
            if obj.attributes.get(self.__config.field) != self._cached:
                self.add_error(
                    object_id=obj.id,
                    detail=f'More than one value detected in {self.__config.field}',
                    field=self.__config.field,
                )



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Iterable

from dateutil.parser import parse as dateutil_parse

from ..core import DataObject, DataSource

JiraTransfer = dict[str, Any]


class Parser(ABC):
    """
    Parses Jira issue `dict`s to `DataObject`
    instances
    """

    def parse_iterable(
        self,
        transfers: Iterable[JiraTransfer]
    ) -> Iterable[DataObject]:
        """
        Parses an `Iterable` of Jira issues
        """

        return (
            self.parse(t) for t in transfers
        )

    @abstractmethod
    def parse(self, transfer: JiraTransfer) -> DataObject:
        """
        Parses an individual Jira issue to a
        `DataObject` instance
        """


class DefaultParser(Parser):
    """
    A trivial implementation of `Parser` that is not intended to be used
    """
    def __init__(
        self,
        data_source_dict: dict[str, DataSource],
        field_mappings: dict[str, dict[str, str]]
    ) -> None:
        self.__dict = data_source_dict
        self.__field_mappings = field_mappings

    __IGNORE_FIELDS = ['attachment', 'comment', 'customfield_10000',
                       'worklog', 'votes', 'watches']

    def parse(self, transfer: JiraTransfer) -> DataObject:
        if 'fields' in transfer:
            return self.__parse_issue(transfer)
        elif 'emailAddress' in transfer:
            return self.__parse_user(transfer)
        elif 'from' in transfer:
            return self.__parse_issue_status_changelog(transfer)

    def __parse_issue(self, transfer: JiraTransfer) -> DataObject:
        type_ = 'issue'
        ds = self.__get_data_source(type_)
        raw_attributes = self.__convert_attributes(type_, transfer['fields'])
        return ds.data_object_factory(
            type_,
            id_=transfer.get('key'),
            attributes={
                k: v for k, v in raw_attributes.items()
                if k not in ['key'] and k in ds.attribute_types[type_]
            } | {
                'status_changes': self.__parse_issue_status_changelog(transfer)
            },
            to_one={
                k: self.__parse_user(v) for k, v in raw_attributes.items()
                if k in ds.relationship_config[type_].to_one
            }
        )

    def __parse_user(self, transfer: JiraTransfer) -> DataObject:
        if transfer is None:
            return None
        if type(transfer) != dict:
            return None
        type_ = 'user'
        ds = self.__get_data_source(type_)
        raw_attributes = transfer
        return ds.data_object_factory(
            type_,
            id_=transfer.get('key'),
            attributes={k: v for k, v in raw_attributes.items()
                        if k not in ['key'] and k in ds.attribute_types[type_]}
        )

    def __parse_issue_status_changelog(self, transfer: JiraTransfer) -> DataObject:
        if transfer is None:
            return None
        status_changes = []
        previous_date = self.__parse_date(transfer['fields']['created'])
        for history in transfer['changelog']['histories']:
            for item in history['items']:
                if item['field'] == 'status':
                    end_date = self.__parse_date(history['created'])
                    status_changes.append({
                        'this_status': item['fromString'],
                        'next_status': item['toString'],
                        'start_date': previous_date,
                        'end_date': end_date
                    })
                    previous_date = end_date
        return status_changes

    def __get_data_source(self, type_: str) -> DataSource:
        return self.__dict[type_]

    def __convert_attributes(
            self,
            type_: str,
            attributes: dict[str, Any]
    ) -> dict[str, Any]:
        ret = {}
        ds = self.__get_data_source(type_)
        for key, value in attributes.items():
            if key in self.__IGNORE_FIELDS:
                continue
            if key in self.__field_mappings:
                ret[self.__field_mappings[key]['system_name']] = \
                    self.__convert_field_value(
                        self.__field_mappings[key]['jira_type'],
                        self.__field_mappings[key]['jira_item_type'],
                        value)
            if type_ in ds.relationship_config and \
                    key in ds.relationship_config[type_].to_one:
                ret[key] = value
        return ret

    def __convert_field_value(self, type_: str, item_type: str, value: Any) -> Any:
        if value is None:
            return None
        elif type_ == 'datetime':
            return self.__parse_date(value)
        elif type_ == 'option':
            return value['value']
        elif type_ == 'array':
            return_value = []
            for array_item in value:
                return_value.append(self.__convert_field_value(item_type, 'string', array_item))
            return return_value
        elif type_ == 'progress':
            return value['progress']
        elif isinstance(value, dict) and 'name' in value:  # Internal Jira types
            return value['name']
        else:
            return value

    def __parse_date(self, date: str) -> str:
        if date is None:
            return None
        return dateutil_parse(date, ignoretz=True)



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class GenomeNotesGenomeNoteToElasticGenomeNoteConverter(
        DataObjectToDataObjectOrUpdateConverter):

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:

        attributes = {
            k: v
            for k, v in data_object.attributes.items()
            if k not in ['taxid', 'tolid', 'assembly_accession']
        }
        to_one_relations = {
            'assembly': self._data_object_factory(
                'assembly',
                data_object.assembly_accession),
            'tolid': self._data_object_factory(
                'tolid',
                data_object.tolid
            ),
            'species': self._data_object_factory(
                'species',
                str(data_object.taxid)
            )
        }
        ret = self._data_object_factory(
            'genome_note',
            data_object.id,
            attributes=attributes,
            to_one=to_one_relations
        )
        yield ret



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os

from ..benchling import (
    BenchlingWarehouseDataSource
)
from ..core import (
    core_data_object
)


def benchling_warehouse(**kwargs) -> BenchlingWarehouseDataSource:
    benchling_warehouse = BenchlingWarehouseDataSource({
        'username': os.getenv('BENCHLING_USER'),
        'password': os.getenv('BENCHLING_PASSWORD'),
        'database': os.getenv('BENCHLING_DB'),
        'hostname': os.getenv('BENCHLING_HOST'),
        'port': os.getenv('BENCHLING_PORT'),
        'schema': os.getenv('BENCHLING_SCHEMA')})
    core_data_object(benchling_warehouse)
    return benchling_warehouse



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Any

from .parser import Parser
from .view import DefaultView
from ..core import DataObject, DataSource, ReqFieldsTree
from ..core.relationship import RelationshipConfig


JsonApiObject = dict[str, Any]
JsonApiTransfer = dict[str, JsonApiObject | list[JsonApiObject]]
JsonRelationship = dict[
    str,  # "one" or "many"
    dict[str, str],  # relationship_name:target_type
]
JsonRelationshipConfig = dict[
    str,  # the object_type
    JsonRelationship,
]


class JsonApiConverter:
    """
    Converts from JSON:API transfers to instances of
    `DataObject`.
    """

    def __init__(
        self,
        parser: Parser,
    ) -> None:
        self.__parser = parser

    def convert(
        self,
        json_xfer: JsonApiTransfer,
    ) -> DataObject:
        """
        Converts a JsonApiTransfer containing a detail (single) result
        """

        obj_list = self.__parser.parse_json_doc(json_xfer)
        return obj_list[0] if obj_list else None

    def convert_list(
        self,
        json_xfer: JsonApiTransfer,
    ) -> tuple[list[DataObject], int | None]:
        """
        Converts a JsonApiTransfer containing a list of results. Also
        returns a count of the total results meeting.
        """

        objs = self.__parser.parse_json_doc(json_xfer)
        total_count = json_xfer.get('meta', {}).get('total', None)
        return objs, total_count

    def convert_count(
        self,
        json_xfer: JsonApiTransfer,
    ) -> dict[str, Any]:
        """
        Converts a JsonApiTransfer containing a list of stats.
        """

        stats = json_xfer['meta']
        return stats['total']

    def convert_stats(
        self,
        json_xfer: JsonApiTransfer,
    ) -> dict[str, Any]:
        """
        Converts a JsonApiTransfer containing a list of stats.
        """

        stats = json_xfer['meta']
        return self.__parser.parse_stats(stats)

    def convert_group_stats(
        self,
        json_xfer: JsonApiTransfer,
    ) -> dict[str, Any]:
        """
        Converts a JsonApiTransfer containing a list of grouped stats.
        """

        stats = json_xfer['meta']
        return self.__parser.parse_group_stats(stats)

    def convert_cursor_page(
        self,
        json_xfer: JsonApiTransfer,
    ) -> tuple[list[DataObject], list[str] | None]:
        """
        Converts a `JsonApiTransfer` of a cursor-page
        """

        objs = self.__parser.parse_json_doc(json_xfer)
        search_after = json_xfer.get('meta', {}).get('search_after')

        return objs, search_after

    def convert_relationship_config(
        self,
        config_transfer: JsonRelationshipConfig,
    ) -> dict[str, RelationshipConfig]:
        """
        Converts a `JsonRelationshipConfig` dict, returned from
        an `api_base2` config blueprint, to a form `ApiDataSource`
        can understand.
        """

        return {
            type_: self.__convert_relationship(rel)
            for type_, rel in config_transfer.items()
        }

    def __convert_relationship(
        self,
        rel: JsonRelationship,
    ) -> RelationshipConfig:
        return RelationshipConfig(
            to_one=rel.get('one'),
            to_many=rel.get('many'),
        )


class DataObjectConverter:
    """
    Converts from instances of `DataObject` to
    JSON:API transfers.
    """

    def __init__(
        self,
        data_source: DataSource,
        prefix: str | None = None,
    ) -> None:
        self.__data_source = data_source
        self.__prefix = prefix

    def __build_view(self, object_type):
        req_fields_tree = ReqFieldsTree(object_type, self.__data_source)
        return DefaultView(req_fields_tree, self.__prefix)

    def convert(self, data_obj: DataObject) -> JsonApiTransfer:
        """
        Converts a single `DataObject` instance to a JsonApiTransfer
        """

        view = self.__build_view(data_obj.type)
        return view.dump(data_obj)

    def convert_list(self, data_obj_list: list[DataObject]) -> JsonApiTransfer:
        """
        Converts a `list` of `DataObject` instances to a JsonApiTransfer
        """

        if not data_obj_list:
            msg = 'Cannot convert empty list'
            raise ValueError(msg)
        view = self.__build_view(data_obj_list[0].type)
        return view.dump_bulk(data_obj_list)



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter,
    ErrorObject
)


class BenchlingEntityToBenchlingWorklistItemConverterFactory():
    def __init__(self, worklist: DataObject):
        self._worklist = worklist

    def get_converter_class(self) -> DataObjectToDataObjectOrUpdateConverter:
        factory = self

        class BenchlingEntityToBenchlingWorklistItemConverter(
                DataObjectToDataObjectOrUpdateConverter):
            def convert(self, data_object: DataObject | ErrorObject) \
                    -> Iterable[DataObject | ErrorObject]:
                if isinstance(data_object, ErrorObject):
                    yield data_object
                else:
                    ret = self._data_object_factory(
                        'worklist_item',
                        to_one={
                            'worklist': factory._worklist,
                            'item': data_object,
                        }
                    )
                    yield ret

        return BenchlingEntityToBenchlingWorklistItemConverter



# SPDX-FileCopyrightText: 2026 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import importlib
from dataclasses import dataclass
from typing import Dict, List, cast

from tol.core import DataObject
from tol.core.validate import ValidationResult, Validator

from .interfaces import Condition, ConditionDict, ConditionEvaluator


Subvalidation = Dict[str, ConditionDict | str | Dict]


class BranchingValidator(Validator, ConditionEvaluator):
    """
    This validator is configured with a list of conditions.
    If a condition passes, the corresponding sub-validator will be run.
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        """
        ```
        validations=[
            {
                'condition': {
                    'field': 'column_name',
                    'operator': '==',
                    'value': 'expected_value',
                },
                'module': '<path.to.module>',
                'class_name': '<path.to.ValidatorClass>',
                'config_details': { ... },
            },
            {
                ...
            }
        ]
        ```
        """
        validations: List[Subvalidation]

    __slots__ = ['__config', '_cached_validators']
    __config: Config
    _cached_validators: Dict[int, Validator]
    """
    Stores all sub-validators that have already been seen so that they can be used again.
    Their keys are their indexes in the `validations` list in the validator config
    """

    def __init__(
        self,
        config: Config,
        **kwargs
    ) -> None:
        super().__init__()

        del kwargs
        self.__config = config
        self._cached_validators = {}

    def _validate_data_object(
        self,
        obj: DataObject
    ) -> None:
        for subvalidator_index, subvalidation in enumerate(self.__config.validations):
            # Do not run this sub-validation if its condition does not pass
            condition_dict = cast(ConditionDict, subvalidation['condition'])
            if not self._does_condition_pass(Condition.from_dict(condition_dict), obj):
                continue

            # Obtain validator, either from cache or anew
            validator: Validator
            if subvalidator_index in self._cached_validators:
                # Use existing validator to perform subvalidation
                validator = self._cached_validators[subvalidator_index]
            else:
                # Create a new validator and use that for the subvalidation
                validator = self.__instantiate_validator(subvalidation)

                # Add the new validator to the store of cached validators
                self._cached_validators[subvalidator_index] = validator

            # Validate data object
            validator._validate_data_object(obj)

            # At this point, I initially added `break` so that once a condition passed, the rest
            # did not need to be checked. However, I decided against it because it is more flexible
            # to have the opportunity to allow multiple conditions to pass, and thus execute
            # multiple subvalidations, on the same `DataObject` in the same iteration

    def __instantiate_validator(self, subvalidation: Subvalidation) -> Validator:
        # Before attempting to extract items from the subvalidation, ensure all of the required
        # keys are there and that they contain values of the correct types.
        # This allows the dictionary to be safely queried thereafter
        try:
            # Accessing with square brackets will also check whether the key exists or not
            if not isinstance(subvalidation['module'], str):
                raise TypeError('module')
            elif not isinstance(subvalidation['class_name'], str):
                raise TypeError('class_name')
        except (KeyError, TypeError) as e:
            # Make errors more specific to validators
            if isinstance(e, KeyError):
                raise KeyError(
                    f'Invalid config in BranchingValidator: `{e.args[0]}` not found'
                )
            else:
                raise TypeError(
                    f'Invalid config in BranchingValidator: '
                    f'`{e.args[0]}` contains an erroneous value type'
                )

        # Dynamically retrieve the validator class
        validator_module = importlib.import_module(subvalidation['module'])
        validator_class = getattr(validator_module, subvalidation['class_name'])

        # Construct and return the new validator
        validator_config = validator_class.Config(
            subvalidation['config_details']
        )
        return validator_class(
            config=validator_config,
        )

    @property
    def results(self) -> List[ValidationResult]:
        """
        Fetches results from all sub-validators, collated into a single list
        """
        return [
            result
            for validator in self._cached_validators.values()
            for result in validator.results
        ]

    @property
    def warnings(self) -> List[ValidationResult]:
        """
        Fetches warnings from all sub-validators, collated into a single list
        """
        return [
            warning
            for validator in self._cached_validators.values()
            for warning in validator.warnings
        ]

    @property
    def errors(self) -> List[ValidationResult]:
        """
        Fetches errors from all sub-validators, collated into a single list
        """
        return [
            error
            for validator in self._cached_validators.values()
            for error in validator.errors
        ]



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import List

from tol.core import Validator
from tol.core.data_object import DataObject


class UniqueWholeOrganismsValidator(Validator):
    """
    Validates an incoming stream of `DataObject` instances.
    For each data object (sample) not a SYMBIONT, it checks:
    1. There are no two samples with organism part WHOLE_ORGANISM with the same SPECIMEN_ID
    2. There are no samples with organism part *not* WHOLE_ORGANISM that have a SPECIMEN_ID
       the same as a WHOLE_ORGANISM in the manifest.
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        symbiont_field: str
        organism_part_field: str
        specimen_id_field: str

    __slots__ = ['__config', '__whole_organisms', '__part_organisms']
    __config: Config
    __whole_organisms: List[str]
    __part_organisms: List[str]

    def __init__(self, config: Config, **kwargs) -> None:
        super().__init__()
        self.__whole_organisms = []
        self.__part_organisms = []
        self.__config = config

    def _validate_data_object(self, obj: DataObject) -> None:
        # This function uses a bit of a confusing method for its validation, so I'm going to
        # leave an explanation here as to how it works for anyone who needs to modify it
        # in the future!
        #
        # In the original code to be adapted, two loops were used. The first looped over each
        # data object whose ORGANISM_PART was 'WHOLE_ORGANISM', adding them to a list. Before it
        # did this though, it would check to see if the SPECIMEN_ID of this data object was already
        # contained in said list (to ensure the specimen IDs were unique). In the second loop, the
        # rest of the data objects (those whose ORGANISM PART was *not* 'WHOLE_ORGANISM') were
        # looped over, each being checked to see if their SPECIMEN_ID was contained in the list
        # (the one containing the whole organisms). In all, this ensures that all whole organisms
        # have unique specimen IDs, and all part organisms do not share the specimen IDs of any of
        # the whole organisms.
        #
        # The issue when adapting this to a Validator to be used in a pipeline, is that this
        # function only takes in one data object at a time, via a generator (to save needing to
        # load many into memory at once). The problem this left us with is that we could no longer
        # achieve the same result by using two passes of the data, as only one pass was feasible.
        #
        # So here's what I ended up with. This validator stores the SPECIMEN_IDs of both all of the
        # whole organisms *and* part organisms. From this, detecting duplicate whole organisms is
        # the same, but detecting whether a part organism shared the SPECIMEN_ID of a whole
        # organism now has two separate cases: when the data object passed into this function is
        # a whole organism, or a part organism. In the case of it being a part organism, a very
        # similar solution to before can be used: we simply check whether self.__whole_organisms
        # conatins the same SPECIMEN_ID. However, for the case where the data object is a whole
        # organism, effectively the inverse is done; it is the self.__part_organisms list that is
        # checked. This covers all cases:
        #   1. There are no duplicates, in which case there will never be a time when the same
        #      SPECIMEN_ID will be in both lists.
        #   2. A whole organism is checked, then a part organism with the same SPECIMEN_ID is
        #      checked. In this case, self.__whole_organisms will contain the same SPECIMEN_ID,
        #      so the duplicate is detected.
        #   3. A part organism is checked, then a whole organism with the same SPECIMEN_ID is
        #      checked. In this case, self.__part_organisms will contain the same SPECIMEN_ID,
        #      so the duplicate is detected.
        #
        # From Thomas :)

        # Ensure the data object is not a SYMBIONT, because organism part checks do not apply
        if obj.attributes.get(self.__config.symbiont_field) != 'SYMBIONT':
            specimen_id = obj.attributes.get(self.__config.specimen_id_field)
            if specimen_id is None:
                return

            organism_part = obj.attributes.get(self.__config.organism_part_field)
            if organism_part == 'WHOLE_ORGANISM':
                if specimen_id in self.__whole_organisms:
                    self.add_error(
                        object_id=obj.id,
                        detail='No two whole organisms can have the same Specimen ID',
                        field=self.__config.specimen_id_field,
                    )
                if specimen_id in self.__part_organisms:
                    self.add_error(
                        object_id=obj.id,
                        detail='A whole organism cannot have a Specimen ID already used for'
                               'a non-whole organism',
                        field=self.__config.specimen_id_field,
                    )

                self.__whole_organisms.append(specimen_id)
            else:
                if specimen_id in self.__whole_organisms:
                    self.add_error(
                        object_id=obj.id,
                        detail='A non-whole organism cannot have a Specimen ID already used for'
                               'a whole organism',
                        field=self.__config.specimen_id_field,
                    )

                self.__part_organisms.append(specimen_id)



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


class JiraMapper():
    def __init__(
        self,
        field_mappings: dict[str, dict[str, str]]
    ) -> None:
        self.__clause_names = self._get_clause_names(field_mappings)

    def _get_clause_names(self, field_mappings: dict[str, dict[str, str]]) -> dict[str, str]:
        all_fields = {
            v['system_name']: v['clause_name']
            for v in field_mappings.values()
        }
        return all_fields | {'id': 'key'}

    def _map_field(self, field: str) -> str:
        field = field.split('.')[0]  # Remove any relation fields
        if field in self.__clause_names:
            return self.__clause_names[field]
        return field



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class GapAssemblyToElasticAssemblyConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert_details(self, assembly: DataObject) -> Iterable[DataObject]:
        return {
            d.id.replace(' ', '_').lower(): d.info
            for d in assembly.assembly_details
        }

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        details = self.convert_details(data_object)

        to_one_relations = {
            'species': self._data_object_factory(
                'species',
                str(data_object.taxon_id)
            ),
        }

        attributes = {
            k: v
            for k, v in data_object.attributes.items()
            if k not in ['taxon_id', 'species', 'phylum_id', 'phylum']
        }

        detail_attributes = {
            k: v
            for k, v in details.items()
            if k in ['organelles', 'total_ingapped_length', 'total_sequence_length',
                     'number_of_chromosomes', 'scaffold_n50', 'number_of_scaffolds',
                     'contig_n50', 'number_of_contigs', 'gc_percent', 'genome_coverage']
        }

        ret = self._data_object_factory(
            'assembly',
            data_object.id,
            attributes={
                **detail_attributes,
                **attributes
            },
            to_one=to_one_relations
        )
        yield ret



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod
from typing import Any

from flask import Flask, request

from ...misc import AuthContext, CtxGetter


class AuthManager(ABC):
    """Manages user profiles and their tokens."""

    @abstractmethod
    def login(self) -> dict[str, str]:
        """The first part of the OIDC handshake."""

    @abstractmethod
    def get_token_from_callback(
        self,
        state: str,
        code: str
    ) -> dict[str, str]:
        """The second part of the OIDC handshake."""

    @abstractmethod
    def create_user_profile(
        self,
        token: str
    ) -> dict[str, Any]:
        """The third part of the OIDC handshake."""

    @abstractmethod
    def revoke_token(
        self,
        token: str
    ) -> None:
        """Deletes the given token."""

    @abstractmethod
    def authenticate(
        self,
        ctx: AuthContext,
        token: str
    ) -> None:
        """
        Sets `str` user_id corresponding to
        the given `token` on the given `ctx`.
        """

    def register(
        self,
        app: Flask,
        ctx_getter: CtxGetter,
        header_name: str = 'token'
    ) -> None:

        @app.before_request
        def __authenticate() -> None:
            if header_name not in request.headers:
                return
            token = request.headers[header_name]

            self.authenticate(
                ctx_getter(),
                token
            )



# SPDX-FileCopyrightText: 2021 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import logging
import signal
import sys

from lab_share_lib.rabbit.background_consumer import BackgroundConsumer
from lab_share_lib.types import RabbitServerDetails

import tol.sciops.configuration as config
from tol.sciops.response_processors import FeedbackProcessor, MessageProcessor

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
# Change the next line to also output pika mq logging
logging.getLogger('pika').propagate = False
LOGGER = logging.getLogger(__name__)


class SciOpsConsumer:
    """ Class that handles async consuming of response messages from Sci Ops """
    @property
    def rabbitmq_details(self):
        return self.__rabbitmq_details

    @property
    def is_healthy(self):
        return self._background_consumer.is_healthy if self._background_consumer else False

    def __init__(self, processor: MessageProcessor):
        """ Constructor """
        LOGGER.info('Initialising SciOps consumer')
        self.__rabbitmq_details = RabbitServerDetails(
            uses_ssl=config.RABBITMQ_USE_SSL,
            host=config.RABBITMQ_HOST,
            port=config.RABBITMQ_PORT,
            username=config.RABBITMQ_USERNAME,
            password=config.RABBITMQ_PASSWORD,
            vhost=config.RABBITMQ_VHOST,
        )
        self.__processor = processor
        self.__background_consumer = None

    def start(self):
        LOGGER.info('Starting SciOps consumer')
        signal.signal(signal.SIGINT, self._signal_handler)
        self.__background_consumer = BackgroundConsumer(
            self.__rabbitmq_details, config.TOL_FEEDBACK_QUEUE, self.__processor.process_message
        )
        # Start and wait for messages
        self.__background_consumer.start()
        self.__background_consumer.join()

    def _signal_handler(self):
        LOGGER.info('Shutting down SciOps consumer')
        if self.__background_consumer:
            self.__background_consumer.stop()


if __name__ == '__main__':
    # TESTING ONLY...
    consumer = SciOpsConsumer(FeedbackProcessor())
    consumer.start()



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from .action import Action  # noqa
from .set_status_action import SetStatusAction  # noqa
from .upsert_action import UpsertAction  # noqa



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Callable

from sqlalchemy import create_engine
from sqlalchemy.orm import Session, sessionmaker


SessionFactory = Callable[[], Session]


def create_session_factory(db_uri: str) -> SessionFactory:
    """
    Creates a Session factory, i.e. a callable that returns a new
    Session object (each time), given the URI of a database.
    """

    engine = create_engine(db_uri,
                           pool_recycle=1800,
                           pool_pre_ping=True,
                           pool_size=2,
                           max_overflow=10)

    session_maker = sessionmaker(
        bind=engine,
        autoflush=True,
        autocommit=False
    )

    return lambda: session_maker()



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import os
import typing

from ..core import (
    core_data_object
)

if typing.TYPE_CHECKING:
    from ..prefect import PrefectDataSource


def prefect(
    insecure: bool = False,
    **kwargs
) -> PrefectDataSource:
    """
    Note - this must be the main entrypoint to the prefect SDK.

    Do not import anything from the `prefect` namespace before running
    this function.
    """

    api_url = os.environ['PREFECT_URL'] + os.environ['PREFECT_API_PATH']

    # this must be set before importing anything from prefect. failure
    # to do so causes the prefect SDK to spin up a local instance.
    os.environ['PREFECT_API_URL'] = api_url

    if insecure:
        os.environ['PREFECT_API_TLS_INSECURE_SKIP_VERIFY'] = '1'

    from ..prefect import create_prefect_datasource

    prefect = create_prefect_datasource(
        api_url,
        insecure=insecure
    )
    core_data_object(prefect)

    return prefect



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import json
from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class ElasticSpeciesToStsSpeciesConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        if data_object is not None:
            legislation = {}
            for att in ['echabs92', 'habreg_2017', 'marhabreg-2017',
                        'waca_1981', 'isb_wildlife_act_1976',
                        'protection_of_badgers_act_1992']:
                att_value = getattr(data_object, f'goat_{att}')
                if att_value is not None:
                    legislation[att] = att_value
            attributes = {
                'legislation': json.dumps(legislation, default=str)
                if legislation != {} else None,
                'family_representative': data_object.goat_family_representative,
                'prefix': data_object.tolid_prefix
            }
            # Don't override the genome size as it may be already set
            if data_object.goat_genome_size is not None:
                attributes['genome_size'] = data_object.goat_genome_size
            yield self._data_object_factory(
                'species',
                data_object.id,
                attributes=attributes
            )



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)
from ...core.operator.updater import DataObjectUpdate


class LabwhereLocationToElasticSampleUpdateConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObjectUpdate]:
        if data_object is not None:
            yield (None, {
                'sts_location': data_object.id,
                'name': data_object.name,
                'parentage': data_object.parentage
            })



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from .parser import Parser  # noqa F401
from .client import JiraClient  # noqa F401
from .converter import JiraConverter  # noqa F401
from .factory import create_jira_datasource  # noqa F401
from .jira_datasource import JiraDataSource  # noqa F401


# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Dict

from ...api_client.exception import BadPostJsonError


class AggregationBody:
    """
    Parses the parameters from a query string for a Aggregation POST
    endpoint.
    """

    def __init__(self, body_dict: Dict) -> None:
        self.__body_dict = body_dict

    @property
    def aggs(self) -> Dict:
        """
        The optional aggregations dict.
        """
        body_dict = self.__body_dict.get('aggs')
        if body_dict is None:
            raise BadPostJsonError(
                'aggs',
                message='"aggs" must be given'
            )
        return body_dict



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Optional

from prefect import get_client
from prefect.client.schemas.filters import (
    DeploymentFilter,
    FlowFilter,
    FlowRunFilter
)
from prefect.settings import (
    PREFECT_API_TLS_INSECURE_SKIP_VERIFY,
    PREFECT_API_URL,
    update_current_profile
)

from .converter import (
    DefaultObjectConverter,
    DefaultPrefectConverter,
    ObjectConverter,
    PrefectConverter
)
from .filter import (
    CLASS_MAP,
    PrefectFilter,
    PrefectFilterBuilder
)
from .prefect_datasource import PrefectDataSource


class _FactoryManager:
    """Manages the converter factory methods"""

    def __init__(self) -> None:
        self.__ds: Optional[PrefectDataSource] = None

    @property
    def data_source(self) -> Optional[PrefectDataSource]:
        return self.__ds

    @data_source.setter
    def data_source(self, prefect_ds: PrefectDataSource) -> None:
        self.__ds = prefect_ds

    def object_converter_factory(self) -> ObjectConverter:
        return DefaultObjectConverter(self.data_source)

    def prefect_converter_factory(self) -> PrefectConverter:
        return DefaultPrefectConverter(self.data_source)


def _filter_factory() -> PrefectFilter:
    builder = PrefectFilterBuilder(
        FlowRunFilter,
        DeploymentFilter,
        FlowFilter,
        CLASS_MAP
    )

    return PrefectFilter(builder)


def _set_config(
    api_url: str,
    insecure: bool
) -> None:

    update_current_profile(
        {
            PREFECT_API_URL: api_url,
        }
    )

    if insecure:
        update_current_profile(
            {
                PREFECT_API_TLS_INSECURE_SKIP_VERIFY: '1'
            }
        )


def create_prefect_datasource(
    api_url: str,
    insecure: bool = False,
) -> PrefectDataSource:
    """Instantiates `PrefectDataSource`."""

    _set_config(api_url, insecure)

    manager = _FactoryManager()

    prefect_ds = PrefectDataSource(
        get_client,
        _filter_factory,
        manager.object_converter_factory,
        manager.prefect_converter_factory
    )

    manager.data_source = prefect_ds

    return prefect_ds



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
# SPDX-License-Identifier: MIT

import datetime
import re
from typing import Iterable

from dateutil import parser as dateutil_parser

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter,
    DataSourceError
)


class StsSampleProjectToElasticSampleConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        # The project (note this is adding to a list)
        s = data_object.sample
        sample_attributes = dict(s.attributes or {})
        if 'storage_rack_id' in sample_attributes:
            sample_attributes['rackid'] = sample_attributes.pop('storage_rack_id')
        attributes = {
            'all_projects': [data_object.project.id],
            'programme': [data_object.project.programme],
            **sample_attributes
        }
        if data_object.project.id and data_object.project.is_primary:
            attributes['project'] = data_object.project.id

        to_one = {}
        try:
            if 'location' in s.to_one_relationships:
                if s.location is not None:
                    location = self.__split_location(s.location.location)
                    if location is not None:
                        attributes['collection_country'] = location['country']
                        attributes['collection_locality'] = location['locality']
                    attributes['latitude'] = s.location.lat
                    attributes['longitude'] = s.location.long
                    attributes['elevation'] = s.location.elevation
                    attributes['depth'] = s.location.depth
                    attributes['habitat'] = s.location.habitat

            attributes['col_date'] = self.__sanitise_date_field(s.col_date)
            attributes['original_collection_date'] = \
                self.__sanitise_date_field(s.original_collection_date)
            attributes['pre_date'] = self.__sanitise_date_field(s.pre_date)
            if 'gal' in s.to_one_relationships:
                if s.gal is not None:
                    attributes['gal_name'] = s.gal.name
                    attributes['gal_abbreviation'] = s.gal.abbreviation
            if 'preservation_approach' in s.to_one_relationships:
                if s.preservation_approach is not None:
                    attributes['preservation_approach'] = s.preservation_approach.approach
            if 'preservative_solution' in s.to_one_relationships:
                if s.preservative_solution is not None:
                    attributes['preservative_solution'] = s.preservative_solution.solution
            if 'collection_method' in s.to_one_relationships:
                if s.collection_method is not None:
                    attributes['collection_method_desc'] = s.collection_method.method
            if 'hazard_group' in s.to_one_relationships:
                if s.hazard_group is not None:
                    attributes['hazard_group'] = s.hazard_group.level
            if 'specimen' in s.to_one_relationships:
                if s.specimen is not None:
                    to_one['specimen'] = self._data_object_factory(
                        'specimen',
                        s.specimen.id
                    )
            if 'sampleset' in s.to_one_relationships:
                if s.sampleset is not None:
                    to_one['sampleset'] = self._data_object_factory(
                        'sampleset',
                        s.sampleset.id
                    )
            if 'manifest' in s.to_one_relationships:
                if s.manifest is not None:
                    to_one['manifest'] = self._data_object_factory(
                        'manifest',
                        s.manifest.id
                    )
            if 'tissue_size' in s.to_one_relationships:
                if s.tissue_size is not None:
                    attributes['tissue_size'] = s.tissue_size.size
            if 'sample_export_options' in s.to_one_relationships:
                if s.sample_export_options is not None:
                    attributes['lab_work_category'] = s.sample_export_options.display_name
            if 'storage_rack' in s.to_one_relationships:
                if s.storage_rack is not None:
                    attributes['location'] = s.storage_rack.freezer_tray.id
            # Make tolid a relationship
            if s.public_name is not None and s.public_name != '':
                to_one['tolid'] = self._data_object_factory(
                    'tolid',
                    s.public_name
                )
                attributes['public_name'] = None

            person_attributes = {}
            for sp in s.sample_persons:
                person_attributes[f'{sp.action.lower()}_name'] = sp.person.fullname

            ext_id_attributes = {}
            for ext_id in s.ext_ids:
                ext_id_attributes[f'{ext_id.ext_id_type.lower()}'] = ext_id.value

            sample_species_attributes = {}
            for ss in s.sample_species:
                if ss.target_or_symbiont == 'TARGET':
                    sample_species_attributes, sample_species_to_one = \
                        self.__convert_sample_species(ss)

        except DataSourceError:
            print(f'Problem with sample {s.id}')

        ret = self._data_object_factory(
            'sample',
            s.id,
            attributes=(
                attributes
                | person_attributes
                | sample_species_attributes
                | ext_id_attributes
            ),
            to_one=to_one | sample_species_to_one
        )
        yield ret

    def __sanitise_date_field(self, date_field: str) -> datetime.datetime:
        if date_field is None or date_field == '':
            return None
        try:
            parsed_date = dateutil_parser.parse(date_field, fuzzy=False)
            return parsed_date
        except ValueError:
            return None

    def __split_location(self, location: str) -> dict[str, str]:
        if location is None:
            return None
        splits = re.split(r'\s*\|\s*', location)
        return {
            'country': splits[0],
            'locality': ' | '.join(splits[1:])
        }

    def __convert_sample_species(self, data_object: DataObject) -> Iterable[DataObject]:
        organism_parts = []
        for ssop in data_object.sample_species_organism_parts:
            organism_parts.append(ssop.organism_part.name)

        return {
            'lifestage':
                data_object.lifestage.name
                if data_object.lifestage is not None else None,
            'strain':
                data_object.strain.name
                if data_object.strain is not None else None,
            'sex':
                data_object.sex.name
                if data_object.sex is not None else None,
            'organism_part': organism_parts
        }, {
            'species': self._data_object_factory(
                'species',
                data_object.species.id
            )
        }



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


import os
from typing import Iterable, Optional

from minio import Minio


class S3Client:

    def __init__(
        self,
        s3_uri: Optional[str] = None,
        access_key: Optional[str] = None,
        secret_key: Optional[str] = None,
        secure: Optional[bool] = None,
    ):
        self.s3_uri = s3_uri or os.environ['S3_URI']
        self.access_key = access_key or os.environ['S3_ACCESS_KEY']
        self.secret_key = secret_key or os.environ['S3_SECRET_KEY']

        if secure is None:
            secure = os.getenv('S3_SECURE', 'true').lower() == 'true'

        self.client = Minio(
            endpoint=self.s3_uri,
            access_key=self.access_key,
            secret_key=self.secret_key,
            secure=secure
        )

    def get_object(
        self,
        bucket_name: str,
        object_name: str,
        file_path: str
    ) -> None:

        self.client.fget_object(bucket_name=bucket_name,
                                object_name=object_name, file_path=file_path)

    def list_objects(
        self,
        bucket_name: str,
        prefix: str = None
    ) -> Iterable:
        yield from self.client.list_objects(bucket_name, prefix=prefix, recursive=True)

    def put_object(
        self,
        bucket_name: str,
        object_name: str,
        file_path: str
    ) -> None:

        self.client.fput_object(bucket_name=bucket_name,
                                object_name=object_name, file_path=file_path)



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import (
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Type,
    TypeVar
)

from ..core.relationship import RelationshipConfig

if typing.TYPE_CHECKING:
    from .sql_converter import TypeFunction
    from .model import Model


class SqlRelationshipConfig(ABC):
    """
    Describes the relationships, to-one and to-many, between
    DataObject instances fulfilled by Model classes within
    SqlDataSource.
    """

    @abstractmethod
    def to_dict(self) -> Optional[Dict[str, RelationshipConfig]]:
        """
        Converts this config to a dictionary representation,
        mapping DataObject types to their individual
        RelationshipConfig instance.
        """


class DefaultSqlRelationshipConfig(ABC):

    def __init__(self, models: List[Type[Model]], type_function: TypeFunction) -> None:
        self.__models = models
        self.__models_dict = {
            m.get_table_name(): m for m in self.__models
        }
        self.__type_function = type_function

    def to_dict(self) -> Dict[str, RelationshipConfig]:

        configs = (
            self.__config_from_model(model) for model in self.__models
        )
        return {
            k: v for k, v in configs if v is not None
        }

    def __config_from_model(self, model: Type[Model]) -> Optional[RelationshipConfig]:

        object_type = self.__type_function(model)
        to_one = model.get_to_one_relationship_config()
        to_many = model.get_to_many_relationship_config()

        if not to_one and not to_many:
            return object_type, None
        else:
            return object_type, RelationshipConfig(
                to_one=self.__map_config(to_one),
                to_many=self.__map_config(to_many),
            )

    def __map_config(self, config: Dict[str, str]) -> Dict[str, str]:

        return {
            k: self.__type_function(self.__models_dict[v])
            for k, v in config.items()
        }


V = TypeVar('V')


class InstanceRelationDict(Mapping, Generic[V], ABC):
    def __init__(self, source: Model) -> None:
        self.__source = source

    def __iter__(self) -> Iterator[str]:
        return iter(self.config)

    def __len__(self) -> int:
        return len(self.config)

    def __getitem__(self, __k: str) -> V:
        if __k not in self.config:
            raise KeyError()
        return getattr(self.source, __k)

    @property
    def source(self) -> Model:
        """The source model for this `dict`"""

        return self.__source

    @property
    @abstractmethod
    def config(self) -> dict[str, str]:
        """
        The relevant relationship config for this `dict`
        """



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class TolqcSpeciesToElasticSpeciesConverter(
        DataObjectToDataObjectOrUpdateConverter):

    ATTRIBUTE_MAPPING = {
        'taxon_family': 'family',
        'taxon_group': 'group',  # change if necessary
        'taxon_order': 'order_group',
        'taxon_phylum': 'taxon_group',
    }

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        target_attributes = {}

        for source_attr, source_value in data_object.attributes.items():
            # if attribute is mapped, use the mapped attribute name
            if source_attr in self.ATTRIBUTE_MAPPING:
                target_attr = self.ATTRIBUTE_MAPPING[source_attr]
            # else, use attribute name as is
            else:
                target_attr = source_attr

            # add attribute to the target dictionary
            target_attributes[target_attr] = source_value

        target_attributes['scientific_name'] = data_object.id
        target_attributes.pop('taxon_id', None)

        # Accessions
        if data_object.data_accession:
            target_attributes['bioproject_accession'] = data_object.data_accession.id
        if data_object.umbrella_accession:
            target_attributes['umbrella_bioproject_accession'] = data_object.umbrella_accession.id

        ret = self._data_object_factory(
            'species',
            data_object.attributes['taxon_id'],
            attributes=target_attributes
        )
        return iter([ret])



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Dict, List, Set

from tol.core import DataObject
from tol.core.validate import Validator


class UniqueValuesValidator(Validator):
    """
    Validates that a stream of `DataObject` instances
    contains unique values for specified keys.
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        unique_keys: List[List[str] | str]
        detail: str = 'Value is not unique'
        is_error: bool = True

    __slots__ = ['__config', '__duplicates', '__existing_values']
    __config: Config
    __duplicates: Dict[str, List[str]]
    __existing_values: Dict[str, Set]

    def __init__(
        self,
        config: Config,
        **kwargs
    ) -> None:

        super().__init__()

        self.__config = config
        self.__duplicates = {}
        self.__existing_values = {}
        for key in self.__config.unique_keys:
            if isinstance(key, str):
                self.__existing_values[key] = set()
            elif isinstance(key, list):
                concat_key = '/'.join(key)
                self.__existing_values[concat_key] = set()

    def _validate_data_object(
        self,
        obj: DataObject
    ) -> None:

        for unique_key in self.__config.unique_keys:
            if isinstance(unique_key, list):
                concat = ''
                for key in unique_key:
                    concat = concat + '/' + (str(obj.attributes[key]))
                if concat in self.__existing_values['/'.join(unique_key)]:
                    self.__duplicate_checks(
                        key=key,
                        value=concat
                    )
                else:
                    self.__existing_values['/'.join(unique_key)].add(concat)

            else:
                if obj.attributes[unique_key] in self.__existing_values[unique_key]:
                    self.__duplicate_checks(
                        key=unique_key,
                        value=obj.attributes[unique_key]
                    )
                else:
                    self.__existing_values[unique_key].add(obj.attributes[unique_key])

    def __duplicate_checks(
        self,
        key: str,
        value: str
    ):
        if key not in self.__duplicates:
            self.__duplicates[key] = []
        self.__duplicates[key].append(value)

    def _post_validation(
        self,
        obj: DataObject,
    ) -> None:
        for key in self.__duplicates:
            self.__add_result(
                obj=obj,
                key=key,
            )

    def __add_result(
        self,
        obj: DataObject,
        key: str,
    ) -> None:

        if self.__config.is_error:
            self.add_error(
                object_id=obj.id,
                detail=self.__config.detail,
                field=key,
            )
        else:
            self.add_warning(
                object_id=obj.id,
                detail=self.__config.detail,
                field=key,
            )



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod

from ..core import DataSourceFilter


class GoatFilter(ABC):
    """
    Converts instances of `DataSourceFilter` to a
    valid GoaT filter.
    """

    @abstractmethod
    def dumps(self, filter_: DataSourceFilter) -> str:
        """
        Emit a filter string from a `DataSourceFilter` instance
        """
        pass


class DefaultGoatFilter(GoatFilter):

    def __init__(
        self
    ) -> None:
        super().__init__()

    __op_mappings = {
        'eq': ['=', '!='],
        'lt': ['<', '>='],
        'lte': ['<=', '>'],
        'gt': ['>', '<='],
        'gte': ['>=', '<'],
        'contains': ['=', '!='],
        'in_list': ['=', '!='],
        'exists': ['is not null', 'is null'],
    }
    __special_fields = {
        'taxon_rank': 'tax_rank',
        'scientific_name': 'tax_name'
    }

    def dumps(self, filter_: DataSourceFilter) -> str:
        filter_conditions = []
        for field, filter_list in filter_.and_.items():
            standard_field = True
            if field == 'id':
                field = 'taxon_id'
            if field in self.__special_fields:
                field = self.__special_fields[field]
                if 'eq' in filter_list and 'value' in filter_list['eq']:
                    filter_conditions.append(f'{field}({filter_list["eq"]["value"]})')
                if 'in_list' in filter_list and 'value' in filter_list['in_list']:
                    filter_conditions.append(
                        f'{field}({",".join(filter_list["in_list"]["value"])})'
                    )
                standard_field = False
            if standard_field:
                for op, ops in self.__op_mappings.items():
                    if op in filter_list:
                        specific_filter = filter_list[op]
                        if 'value' in specific_filter:
                            operator = ops[0]
                            if 'negate' in specific_filter:
                                operator = ops[1]
                            if op == 'in_list':
                                filter_condition = ','.join(
                                    f'{v}' for v in specific_filter['value']
                                )
                            elif op == 'contains':
                                filter_condition = str(specific_filter['value']) + '*' \
                                    if 'value' in specific_filter else ''
                            else:
                                filter_condition = str(specific_filter['value']) \
                                    if 'value' in specific_filter else ''
                            filter_conditions.append(f'{field}{operator}'
                                                     f'{filter_condition}')
        filter_string = ' AND '.join(filter_conditions)
        return filter_string if filter_string is not None else ''



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Iterable, Optional

from ._writer import _Writer

if typing.TYPE_CHECKING:
    from ..data_object import ErrorObject
    from ..session import OperableSession


class Deleter(_Writer, ABC):
    """
    Deletes DataObject instances.
    """

    @abstractmethod
    def delete(
        self,
        object_type: str,
        object_ids: Iterable[str],
        session: Optional[OperableSession] = None
    ) -> Iterable[ErrorObject | None] | None:
        """
        Takes a type, and the IDs of the `DataObject` to delete
        of the stated type.
        """



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import tempfile
import uuid
import xml.etree.ElementTree as ElementTree
from typing import Dict, List, Tuple

sample_xml_template = """<?xml version="1.0" ?>
<SAMPLE_SET xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation=\
"ftp://ftp.sra.ebi.ac.uk/meta/xsd/sra_1_5/SRA.sample.xsd">
</SAMPLE_SET>"""


submission_xml_template = """<?xml version="1.0" encoding="UTF-8"?>
<SUBMISSION xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation=\
"ftp://ftp.sra.ebi.ac.uk/meta/xsd/sra_1_5/SRA.submission.xsd">
<CONTACTS></CONTACTS>
<ACTIONS>
<ACTION>
<ADD/>
</ACTION>
<ACTION>
<RELEASE/>
</ACTION>
</ACTIONS>
</SUBMISSION>"""


def convert_checklist_xml_to_dict(checklist_xml: str) -> Dict[str, Tuple[str, str, object]]:
    # key label, val [mandatory_status, ]
    checklists = []
    root = ElementTree.fromstring(checklist_xml)
    for checklist_node in root.findall('./CHECKLIST'):
        checklist_id = checklist_node.get('accession')
        fields = {}
        for field_group_node in checklist_node.findall('./DESCRIPTOR/FIELD_GROUP'):
            for field_node in field_group_node.findall('./FIELD'):

                label, mandatory_status = None, None

                label_node = field_node.find('./LABEL')

                if label_node is not None:
                    label = label_node.text

                mandatory_node = field_node.find('./MANDATORY')

                if mandatory_node is not None:
                    mandatory_status = mandatory_node.text

                regex_node = field_node.find('./FIELD_TYPE/TEXT_FIELD/REGEX_VALUE')
                if regex_node is not None:
                    regex_str = regex_node.text
                    fields[label] = [mandatory_status, 'restricted text', regex_str]
                    continue

                text_choice_node = field_node.find('./FIELD_TYPE/TEXT_CHOICE_FIELD')

                if text_choice_node is not None:
                    text_options = []
                    for text_option_node in text_choice_node.findall('./TEXT_VALUE/VALUE'):
                        text_options.append(text_option_node.text)

                    fields[label] = [mandatory_status, 'text choice', text_options]
                    continue

                taxon_node = field_node.find('./FIELD_TYPE/TEXT_FIELD/TAXON_FIELD')

                if taxon_node is not None:
                    regex_str = regex_node.text
                    fields[label] = [mandatory_status, 'valid taxonomy', '']
                    continue

                fields[label] = [mandatory_status, 'free text', '']
        checklists.append({'checklist_id': checklist_id, 'checklist': fields})
    return checklists


def convert_xml_to_list_of_sample_dict(response_xml: str) -> List[Dict[str, List[str]]]:
    samples = []
    # Convert sample xml to dictionary
    # SAMPLE_ATTRIBUTE use TAG as key, tuple (VALUE, UNITS)
    # Additional entries TITLE, SAMPLE_NAME, TAXONID

    root = ElementTree.fromstring(response_xml)
    for xml_sample_node in root.findall('./SAMPLE'):
        sample = {}

        title, taxon_id, scientific_name = None, None, None

        title_node = xml_sample_node.find('./TITLE')
        taxon_id_node = xml_sample_node.find('./SAMPLE_NAME/TAXON_ID')
        scientific_name_node = xml_sample_node.find('./SAMPLE_NAME/SCIENTIFIC_NAME')

        if title_node is not None:
            title = title_node.text

        if taxon_id_node is not None:
            taxon_id = taxon_id_node.text

        if scientific_name_node is not None:
            scientific_name = scientific_name_node.text

        sample['title'] = [title, None]
        sample['taxon_id'] = [taxon_id, None]
        sample['scientific_name'] = [scientific_name, None]

        for xml_sample_attr_node in \
                xml_sample_node.findall('./SAMPLE_ATTRIBUTES/SAMPLE_ATTRIBUTE'):
            tag, val, units = None, None, None

            tag_node = xml_sample_attr_node.find('./TAG')
            val_node = xml_sample_attr_node.find('./VALUE')
            units_node = xml_sample_attr_node.find('./UNITS')

            if tag_node is not None:
                tag = tag_node.text

            if val_node is not None:
                val = val_node.text

            if units_node is not None:
                units = units_node.text

            sample[tag] = [val, units]

        samples.append(sample)

    return samples


def build_bundle_sample_xml(samples: Dict[str, Dict[str, List[str]]]) -> Tuple[str, int]:
    """build structure and save to file bundle_file_subfix.xml"""

    manifest_id = uuid.uuid4()

    dir_ = tempfile.TemporaryDirectory()

    filename = f'{dir_.name}bundle_{str(manifest_id)}.xml'

    with open(filename, 'w') as sample_xml_file:
        sample_xml_file.write(sample_xml_template)

    sample_count = update_bundle_sample_xml(samples, filename)

    return filename, sample_count


def update_bundle_sample_xml(samples: Dict[str, Dict[str, List[str]]], bundlefile: str) -> int:
    """update the sample with submission alias adding a new sample"""

    # print('adding sample to bundle sample xml')
    tree = ElementTree.parse(bundlefile)
    root = tree.getroot()
    sample_count = 0
    for title, sample in samples.items():
        sample_count += 1
        sample_alias = ElementTree.SubElement(root, 'SAMPLE')

        # Title is format <unique id>-<project name>-<specimen_type>
        t_arr = title.split('-')

        sample_alias.set('alias',
                         f'{t_arr[0]}-{t_arr[1]}-{t_arr[2]}-{t_arr[3]}-{t_arr[4]}')
        sample_alias.set('center_name', 'SangerInstitute')

        title_block = ElementTree.SubElement(sample_alias, 'TITLE')
        title_block.text = title
        sample_name = ElementTree.SubElement(sample_alias, 'SAMPLE_NAME')
        taxon_id = ElementTree.SubElement(sample_name, 'TAXON_ID')
        taxon_id.text = str(sample['taxon_id'][0])
        scientific_name = ElementTree.SubElement(sample_name, 'SCIENTIFIC_NAME')
        scientific_name.text = str(sample['scientific_name'][0])
        sample_attributes = ElementTree.SubElement(sample_alias, 'SAMPLE_ATTRIBUTES')

        for key, val in sample.items():

            if key in ['title', 'taxon_id', 'scientific_name']:
                continue

            sample_attribute = ElementTree.SubElement(sample_attributes, 'SAMPLE_ATTRIBUTE')
            tag = ElementTree.SubElement(sample_attribute, 'TAG')
            tag.text = key
            value = ElementTree.SubElement(sample_attribute, 'VALUE')
            value.text = str(val[0])
            # add ena units where necessary
            if val[1]:
                unit = ElementTree.SubElement(sample_attribute, 'UNITS')
                unit.text = val[1]

    ElementTree.dump(tree)
    tree.write(open(bundlefile, 'w'),
               encoding='unicode')
    return sample_count


def build_submission_xml(manifest_id: str, contact_name: str, contact_email: str) -> str:

    dir_ = tempfile.TemporaryDirectory()

    submissionfile = f'{dir_.name}submission_{str(manifest_id)}.xml'

    with open(submissionfile, 'w') as submission_xml_file:
        submission_xml_file.write(submission_xml_template)

    # build submission XML
    tree = ElementTree.parse(submissionfile)
    root = tree.getroot()

    # set SRA contacts
    contacts = root.find('CONTACTS')

    # set copo sra contacts
    copo_contact = ElementTree.SubElement(contacts, 'CONTACT')
    copo_contact.set('name', contact_name)
    copo_contact.set('inform_on_error', contact_email)
    copo_contact.set('inform_on_status', contact_email)
    ElementTree.dump(tree)

    tree.write(open(submissionfile, 'w'),
               encoding='unicode')

    return submissionfile


def assign_ena_ids(samples: str, xml: str) -> Dict[str, Dict[str, List[str]]]:

    try:
        tree = ElementTree.fromstring(xml)
    except ElementTree.ParseError:
        return False

    success_status = tree.get('success')
    if success_status == 'false':
        return False
    else:
        return assign_biosample_accessions(samples, xml)


def assign_biosample_accessions(samples: Dict[str, Dict[str, List[str]]],
                                xml: str) -> Dict[str, Dict[str, List[str]]]:
    # Parse response to return generated biosample ids

    assigned_samples = {}

    tree = ElementTree.fromstring(xml)
    submission_accession = tree.find('SUBMISSION').get('accession')
    for child in tree.iter():
        if child.tag == 'SAMPLE':
            sample_id = child.get('alias')
            sra_accession = child.get('accession')
            biosample_accession = child.find('EXT_ID').get('accession')

            for key, sample_dict in samples.items():

                if sample_id in key:
                    sample_dict['sra_accession'] = [sra_accession, None]
                    sample_dict['biosample_accession'] = [biosample_accession, None]
                    sample_dict['submission_accession'] = [submission_accession, None]

                    assigned_samples[key] = sample_dict

    return assigned_samples



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import json
from typing import Dict, Iterable, Optional
from urllib.parse import quote

from .converter import BoldApiTransfer
from ..core import HttpClient


class BoldApiClient(HttpClient):
    """
    Takes BOLD API transfers and connects to a remote
    BOLD API.
    """

    def __init__(
        self,
        bold_url: str,
        bold_portal_url: str,
        bold_api_key: str,
        retries: int = 5
    ) -> None:
        super().__init__(token=bold_api_key, token_header='api-key', retries=retries)
        self.__bold_url = bold_url
        self.__bold_portal_url = bold_portal_url

    def get_detail(
        self,
        object_type: str,
        object_ids: Iterable[str]
    ) -> Optional[BoldApiTransfer]:
        """
        Gets a list of BOLD API transfers for the objects of specified
        `object_type` and `object_id`, or returns None if not found.
        """
        if object_type == 'bin':
            return (
                self.__get_detail_portal(object_type, object_id)
                for object_id in object_ids
            )
        else:
            return self.__get_detail_data(object_type, object_ids)

    def __get_detail_data(
        self,
        object_type: str,
        object_ids: Iterable[str]
    ) -> Optional[BoldApiTransfer]:
        """
        Gets a list of BOLD API transfers for the objects of specified
        `object_type` and `object_id`, or returns None if not found.
        """

        url, params = self.__detail_url(object_type, object_ids)
        headers = self._merge_headers()
        return self.__fetch_detail(url, params=params, headers=headers)

    def __get_detail_portal(
        self,
        object_type: str,
        object_id: str
    ) -> Optional[BoldApiTransfer]:
        """
        Gets a list of BOLD API transfers for the objects of specified
        `object_type` and `object_id`, or returns None if not found.
        """

        url, params = self.__query_url(object_type, object_id)
        headers = self._merge_headers()
        query_id = self.__fetch_query(url, params=params, headers=headers)
        if not query_id:
            return None
        return self.__fetch_bin(self.__bin_url(query_id), object_id, headers=headers)

    def __fetch_detail(
        self,
        url: str,
        params: Dict = {},
        headers: Dict = {},
    ) -> Optional[BoldApiTransfer]:

        session = self._get_session_with_retries()
        r = session.get(url, params=params, headers=headers)
        if r.status_code in [400, 404]:
            return []
        r.raise_for_status()
        results = []
        lines = r.text.splitlines()
        for line in lines:
            result = json.loads(line)
            results.append(result)
        return results

    def __detail_url(self, object_type: str, object_ids: str) -> str:
        url = f'{self.__bold_url}/records'
        obj_ids_str = ','.join(object_ids)
        params = {
            'sampleids': obj_ids_str
        }
        return url, params

    def __query_url(self, object_type: str, object_id: str) -> str:
        url = f'{self.__bold_portal_url}/query'
        params = {
            'query': f'{object_type}:uri:{object_id}'
        }
        return url, params

    def __fetch_query(
        self,
        url: str,
        params: Dict = {},
        headers: Dict = {},
    ) -> Optional[BoldApiTransfer]:

        session = self._get_session_with_retries()
        r = session.get(url, params=params, headers=headers)
        if r.status_code in [400, 404]:
            return None
        r.raise_for_status()
        results = r.json()
        return results['query_id'] if 'query_id' in results else None

    def __bin_url(self, query_id: str) -> str:
        url = f'{self.__bold_portal_url}/taxonomy/{quote(query_id)}'
        return url

    def __fetch_bin(
        self,
        url: str,
        object_id: str,
        params: Dict = {},
        headers: Dict = {},
    ) -> Optional[BoldApiTransfer]:

        session = self._get_session_with_retries()
        r = session.get(url, params=params, headers=headers)
        if r.status_code in [400, 404]:
            return None
        r.raise_for_status()
        results = r.json()
        # Unmatched queries just return empty dicts
        return results | {'binid': object_id} \
            if 'taxonomy' in results \
            and 'kingdom' in results['taxonomy'] \
            and results['taxonomy']['kingdom'] != {} \
            else None



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class GapAssemblyToElasticAssemblyAnalysisConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert_details(self, assembly: DataObject) -> Iterable[DataObject]:
        return {
            d.id.replace(' ', '_').lower(): d.info
            for d in assembly.assembly_details
        }

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:

        for p in data_object.pipelines:
            ret = self._data_object_factory(
                'assembly_analysis',
                f'{data_object.id}_{p.id}',
                attributes={
                    'analysis': p.analysis,
                    'results': p.results,
                    's3': p.s3,
                    'lustre_path_analysis': p.lustre_path_analysis,
                },
                to_one={
                    'assembly': self._data_object_factory(
                        'assembly',
                        data_object.id
                    ),
                    'species': self._data_object_factory(
                        'species',
                        str(data_object.taxon_id)
                    ),
                }
            )
            yield ret



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import Any

from dateutil.parser import parse as dateutil_parse

from ..core import DataObject, ReqFieldsTree

if typing.TYPE_CHECKING:
    from ..core import DataSource


JsonApiResource = dict[str, Any]
JsonApiDoc = dict[str, list[JsonApiResource]]


class Parser(ABC):
    """
    Parses JSON:API transfer resource `dict`s to `DataObject`
    instances
    """

    @abstractmethod
    def parse_json_doc(self, transfer: JsonApiDoc) -> Iterable[DataObject]:
        """
        Parses a JSON:API document, which includes a `data` array and possibly
        an `included` array of related objects, returning an list of
        `DataObject`.
        """

    @abstractmethod
    def parse_stats(self, transfer: JsonApiResource) -> dict:
        """
        Parses an individual stats transfer resource to a
        stats dict instance
        """

    @abstractmethod
    def parse_group_stats(self, transfer: JsonApiResource) -> list[dict]:
        """
        Parses a grouped stats transfer resource to a
        list instance
        """


class DefaultParser(Parser):
    def __init__(
        self,
        data_source_dict: dict[str, DataSource],
        requested_tree: ReqFieldsTree | None = None,
    ) -> None:
        self.__ds_dict = data_source_dict
        self.__requested_tree = requested_tree

    def parse_json_doc(
        self,
        transfer: JsonApiDoc,
    ) -> Iterable[DataObject]:
        data_objects = list(self.__parse_iterable(transfer['data']))
        if tree := self.__requested_tree:
            included = DataObjectCatalog(self.__parse_iterable(transfer.get('included')))
            for obj in data_objects:
                self.__link_related_obejcts(tree, included, obj)
        return data_objects

    def __link_related_obejcts(
        self,
        tree: ReqFieldsTree,
        included: DataObjectCatalog,
        data_object: DataObject,
    ) -> None:
        """
        Using the `ReqFieldsTree` recursively replaces related stub
        `DataObject`s with `DataObject`s from the `incldued`
        `DataObjectCatalog` which were built from the JSON:API "included"
        array.
        """
        for name, sub_tree in tree.sub_trees():
            if name in tree.to_one_names():
                if (related := data_object._to_one_objects.get(name)) and (
                    inc := included.fetch(related)
                ):
                    # Link the to-one object
                    setattr(data_object, name, inc)
                    self.__link_related_obejcts(sub_tree, included, inc)
            elif related := data_object._to_many_objects.get(name):
                # Link each to-many object
                for i, rel in enumerate(related):
                    if inc := included.fetch(rel):
                        related[i] = inc
                        self.__link_related_obejcts(sub_tree, included, inc)

    def __parse_iterable(
        self,
        transfer: list[JsonApiResource],
    ) -> Iterable[DataObject]:
        if transfer:
            if isinstance(transfer, list):
                for json_res in transfer:
                    yield self.__parse(json_res)
            else:
                yield self.__parse(transfer)

    def __parse(self, transfer: JsonApiResource) -> DataObject:
        type_ = transfer['type']
        ds = self.__get_data_source(type_)
        attributes = self.__convert_attributes(type_, transfer.get('attributes'))
        to_one, to_many = self.__parse_relationships(transfer.get('relationships'))

        return ds.data_object_factory(
            type_,
            id_=transfer.get('id'),
            attributes=attributes,
            to_one=to_one,
            to_many=to_many,
        )

    def parse_stats(self, transfer: JsonApiResource) -> dict:
        type_ = transfer.get('type')
        raw_stats = transfer.get('stats')
        converted_stats = self.__convert_stats(type_, raw_stats)
        return {'stats': converted_stats}

    def parse_group_stats(self, transfer: JsonApiResource) -> Iterable[dict]:
        type_ = transfer.get('type')
        raw_stats = transfer.get('stats')

        return [self.__convert_group_stats(type_, raw_stat) for raw_stat in raw_stats]

    def __get_data_source(self, type_: str) -> DataSource:
        return self.__ds_dict[type_]

    def __parse_relationships(
        self, related: dict[str, JsonApiResource] | None
    ) -> tuple[dict[str, DataObject | None], dict[str, list[DataObject]]]:
        to_one = {}
        to_many = {}
        if related:
            for name, value in related.items():
                if value is None:
                    # This must be a to-one relation because to-many relations
                    # are never null.  (If the to-many has been fetched it
                    # will be an empty list. If it has not been fetched it
                    # will be a dict containing a "links" key.)
                    to_one[name] = None
                elif data := value.get('data'):
                    if isinstance(data, list):
                        to_many[name] = [self.__make_stub_data_object(x) for x in data]
                    else:
                        to_one[name] = None if data is None else self.__make_stub_data_object(data)
        return to_one, to_many

    def __make_stub_data_object(self, transfer: JsonApiResource):
        type_ = transfer['type']
        ds = self.__get_data_source(type_)
        return ds.data_object_factory(
            type_,
            id_=transfer['id'],
            stub=True,
        )

    def __convert_attributes(
        self, type_: str, attributes: dict[str, Any] | None
    ) -> dict[str, Any]:
        if not attributes:
            return {}

        datetime_keys = self.__get_datetime_keys(type_)

        return {
            k: (dateutil_parse(v) if k in datetime_keys and v is not None else v)
            for k, v in attributes.items()
        }

    def __convert_stats(self, type_: str, stats: dict[str, Any] | None) -> dict[str, Any]:
        # {'field': {'min': value, 'max': value}
        if not stats:
            return {}

        datetime_keys = self.__get_datetime_keys(type_)

        return {
            fieldname: {
                k: (
                    dateutil_parse(v, ignoretz=True)
                    if fieldname in datetime_keys and v is not None and k in ['min', 'max']
                    else v
                )
                for k, v in fieldstats.items()
            }
            for fieldname, fieldstats in stats.items()
        }

    def __convert_group_stats(
        self, type_: str, raw_stats: dict[str, dict[str, Any]]
    ) -> dict[str, dict[str, Any]]:
        st = raw_stats.pop('stats')
        count = st.pop('count', None)

        raw_stats['stats'] = self.__convert_stats(type_, st)

        if count is not None:
            raw_stats['stats']['count'] = count

        return raw_stats

    def __get_datetime_keys(self, type_: str) -> set[str]:
        """
        Gets called on each object, which is somewhat inefficient.  Should be
        cached for each object type, but don't want `self` in a cache because
        it could leak memory.
        """
        ds = self.__get_data_source(type_)
        attribute_types = ds.attribute_types.get(type_, {})

        return {attr for attr, typ in attribute_types.items() if self.__type_is_datetime(typ)}

    def __type_is_datetime(self, typ: str, /) -> bool:
        lc_type = typ.lower()

        return 'date' in lc_type or 'time' in lc_type


class DataObjectCatalog:
    """
    A catalog of `DataObject`s keyed by their `type` and `id` attributes.
    """

    def __init__(self, data_obj_list: Iterable[DataObject] | None):
        self.__obj_index = {}
        if data_obj_list:
            for obj in data_obj_list:
                self.store(obj)

    def __len__(self):
        return len(self.__obj_index)

    def store(self, obj) -> None:
        key = obj.type, obj.id
        self.__obj_index[key] = obj

    def fetch(self, obj) -> DataObject | None:
        key = obj.type, obj.id
        return self.__obj_index.get(key)



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class BenchlingExtractionToElasticExtractionConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        if data_object.sts_id and data_object.id:
            ret = self._data_object_factory(
                'extraction',
                data_object.id,
                attributes={
                    **{k: v
                       for k, v in data_object.attributes.items()
                       if k not in ['sts_id', 'specimen_id', 'taxon_id',
                                    'programme_id', 'eln_tissue_prep_id']}
                },
                to_one={
                    'sample': self._data_object_factory(
                        'sample',
                        data_object.sts_id
                    ) if data_object.sts_id is not None else None,
                    'species': self._data_object_factory(
                        'species',
                        data_object.taxon_id
                    ) if data_object.taxon_id is not None else None,
                    'specimen': self._data_object_factory(
                        'specimen',
                        data_object.specimen_id
                    ) if data_object.specimen_id is not None else None,
                    'tolid': self._data_object_factory(
                        'tolid',
                        data_object.programme_id
                    ) if data_object.programme_id is not None else None,
                    'tissue_prep': self._data_object_factory(
                        'tissue_prep',
                        data_object.eln_tissue_prep_id
                    ) if data_object.eln_tissue_prep_id is not None else None,
                }
            )
            yield ret



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from .csv_datasource import CsvDataSource  # noqa
from .excel import *  # noqa
from .excel_datasource import ExcelDataSource  # noqa
from .google_factory import google_csv_datasource_factory  # noqa
from .s3_factory import s3_excel_datasource_factory  # noqa



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from collections.abc import Mapping
from typing import Callable, Iterator, Optional

from .client import GoatApiClient
from .converter import (
    GoatApiConverter
)
from .filter import DefaultGoatFilter
from .goat_datasource import (
    GoatConverterFactory,
    GoatDataSource
)
from .parser import DefaultParser
from ..core import DataSource


class _GoatDSDict(Mapping):
    def __init__(self, api_ds: GoatDataSource) -> None:
        self.__ds = api_ds

    def __getitem__(self, __k: str) -> GoatDataSource:
        if __k not in self.__ds.supported_types:
            raise KeyError()
        return self.__ds

    def __iter__(self) -> Iterator[str]:
        return iter(self.__ds.supported_types)

    def __len__(self) -> int:
        return len(self.__ds.supported_types)


class _ConverterFactory:
    """
    Manages the instantation of:

    - `GoatApiConverter`
    """

    def __init__(self) -> None:
        self.__data_source: Optional[DataSource] = None

    @property
    def data_source(self) -> Optional[DataSource]:
        return self.__data_source

    @data_source.setter
    def data_source(
        self,
        ds: DataSource
    ) -> None:

        self.__data_source = ds

    def goat_converter_factory(self) -> GoatConverterFactory:
        """
        Returns an instantiated `GoatApiConverter`.
        """

        parser = DefaultParser(self.__ds_dict)
        return GoatApiConverter(parser)

    @property
    def __ds_dict(self) -> dict[str, DataSource]:
        return _GoatDSDict(self.data_source)


class _FilterFactory:
    """
    Manges the instantation of:

    - `GoatFilter`
    """

    def __init__(self) -> None:
        pass

    def goat_filter_factory(self) -> DefaultGoatFilter:
        """
        Returns an instantiated `GoatFilter`.
        """

        return DefaultGoatFilter()


def _get_client_factory(
    api_url: str
) -> Callable[[], GoatApiClient]:
    """
    A resonable default for creating
    a `GoatApiClient` instance
    """

    return lambda: GoatApiClient(
        api_url
    )


def create_goat_datasource(
    goat_url: str
) -> GoatDataSource:
    """
    Instantiates `GoatDataSource` using the given:

    - `goat_url`
    """

    client_factory = _get_client_factory(
        goat_url
    )
    manager = _ConverterFactory()
    filter_factory = _FilterFactory()
    goat_ds = GoatDataSource(
        client_factory,
        manager.goat_converter_factory,
        filter_factory.goat_filter_factory
    )

    manager.data_source = goat_ds

    return goat_ds



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import (
    Iterable
)

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class ElasticObjectToPortaldbObjectConverter(
        DataObjectToDataObjectOrUpdateConverter):

    def __init__(self, data_object_factory, fields: dict = {},
                 destination_object_type: str = 'tolid_event',
                 id_field: str = 'id', incremental: bool = False):
        super().__init__(data_object_factory)
        self.__fields = fields
        self.__destination_object_type = destination_object_type
        self.__id_field = id_field
        self.__incremental = incremental

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        if data_object is not None:

            ret = self._data_object_factory(
                self.__destination_object_type,
                data_object.get_field_by_name(self.__id_field),
                attributes=self.__fields
            )

            if self.__incremental and self.__destination_object_type == 'tolid_event':
                obj = ret._host.get_one(
                    'tolid_event',
                    data_object.get_field_by_name(self.__id_field)
                )
                count_field_value = obj.tol_tum_action_count if obj else 0
                new_count = count_field_value + 1 if count_field_value else 1
                ret.attributes['tol_tum_action_count'] = new_count if self.__incremental else None

            yield ret
        else:
            yield None



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from collections.abc import Mapping
from typing import Callable, Dict, Iterator, Optional

from .client import EnaApiClient
from .converter import (
    EnaApiConverter
)
from .ena_datasource import (
    EnaConverterFactory,
    EnaDataSource
)
from .filter import DefaultEnaFilter
from .parser import DefaultParser
from ..core import DataSource


class _EnaDSDict(Mapping):
    def __init__(self, api_ds: EnaDataSource) -> None:
        self.__ds = api_ds

    def __getitem__(self, __k: str) -> EnaDataSource:
        if __k not in self.__ds.supported_types:
            raise KeyError()
        return self.__ds

    def __iter__(self) -> Iterator[str]:
        return iter(self.__ds.supported_types)

    def __len__(self) -> int:
        return len(self.__ds.supported_types)


class _ConverterFactory:
    """
    Manages the instantiation of

    - `EnaApiConverter`
    """

    def __init__(self) -> None:
        self.__data_source: Optional[DataSource] = None

    @property
    def data_source(self) -> Optional[DataSource]:
        return self.__data_source

    @data_source.setter
    def data_source(
        self,
        ds: DataSource
    ) -> None:

        self.__data_source = ds

    def ena_converter_factory(self) -> EnaConverterFactory:
        """
        Returns an instantiated `EnaApiConverter`.
        """
        parser = DefaultParser(self.__ds_dict)
        return EnaApiConverter(parser)

    @property
    def __ds_dict(self) -> dict[str, DataSource]:
        return _EnaDSDict(self.data_source)


def _get_client_factory(
    ena_url: str,
    ena_user: str,
    ena_password: str,
    ena_contact_name: str,
    ena_contact_email: str
) -> Callable[[], EnaApiClient]:
    """
    A resonable default for creating an `EnaApiClient` instance.
    """
    return lambda: EnaApiClient(
        ena_url,
        ena_user,
        ena_password,
        ena_contact_name,
        ena_contact_email
    )


class _FilterFactory:
    """
    Manages the instantiation of

    - `EnaFilter`
    """

    def __init__(self) -> None:
        pass

    def ena_filter_factory(self) -> DefaultEnaFilter:
        """
        Returns an instantiated `EnaFilter`.
        """

        return DefaultEnaFilter()


def create_ena_datasource(
    ena_url: str,
    ena_user: str,
    ena_password: str,
    ena_contact_name: str,
    ena_contact_email: str
) -> EnaDataSource:
    """
    Instantiates `EnaDataSource` using the given:

    - `ena_url`
    - `ena_user`
    - `ena_password`
    - `ena_contact_name`
    - `ena_contact_email`
    """

    client_factory = _get_client_factory(
        ena_url,
        ena_user,
        ena_password,
        ena_contact_name,
        ena_contact_email
    )
    manager = _ConverterFactory()
    filter_factory = _FilterFactory()

    ena_ds = EnaDataSource(
        client_factory,
        manager.ena_converter_factory,
        filter_factory.ena_filter_factory
    )

    manager.data_source = ena_ds

    return ena_ds


def create_ena_datasource_from_config(
    ena_url: str,
    config: Dict
) -> EnaDataSource:
    """
    Instantiates `EnaDataSource` using the given `config`.
    """

    return create_ena_datasource(
        ena_url,
        config['user'],
        config['password'],
        config['contact_name'],
        config['contact_email']
    )



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os

from .defaults import Defaults
from ..api_client import (
    ApiDataSource,
    create_api_datasource
)
from ..core import (
    core_data_object
)


def portal(
        retries: int = 5,
        dataspace: str = 'tol_production',
        **kwargs
) -> ApiDataSource:
    portal = create_api_datasource(
        api_url=os.getenv('PORTAL_URL', Defaults.PORTAL_URL)
        + os.getenv('PORTAL_API_PATH', Defaults.PORTAL_API_PATH),
        token=os.getenv('PORTAL_API_KEY'),
        data_prefix=os.getenv('PORTAL_API_DATA_PATH', Defaults.PORTAL_API_DATA_PATH)
        + f'/{dataspace}',
        retries=retries
    )
    core_data_object(portal)
    return portal



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os

from benchling_sdk.auth.api_key_auth import ApiKeyAuth
from benchling_sdk.benchling import Benchling

ELN_URL = os.getenv('ELN_URL')
ELN_API_KEY = os.getenv('ELN_API_KEY')


def get_benchling_instance():
    return Benchling(
        url=ELN_URL,
        auth_method=ApiKeyAuth(ELN_API_KEY)
    )



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

def generate_flattened_type(type_, **kwargs):
    """A generic method for getting a 'flat' generator of benchling
    types (i.e. one by one) from a list endpoint, e.g. assay_results
    or containers."""
    generator = type_.list(**kwargs)
    for page in generator:
        yield from page


def generate_workflow_tasks(benchling, **kwargs):
    return generate_flattened_type(
        benchling.workflow_tasks,
        **kwargs
    )


def generate_workflow_outputs(benchling, **kwargs):
    return generate_flattened_type(
        benchling.workflow_outputs,
        **kwargs
    )


def generate_assay_results(benchling, **kwargs):
    return generate_flattened_type(
        benchling.assay_results,
        **kwargs
    )


def generate_containers(benchling, **kwargs):
    return generate_flattened_type(
        benchling.containers,
        **kwargs
    )


def generate_boxes(benchling, **kwargs):
    return generate_flattened_type(
        benchling.boxes,
        **kwargs
    )


def generate_custom_entities(benchling, **kwargs):
    return generate_flattened_type(
        benchling.custom_entities,
        **kwargs
    )
